import unittest
from deepchem.utils import handle_hydrogen
from deepchem.utils import make_polymer_mol
from deepchem.utils import parse_polymer_rules
from deepchem.utils import tag_atoms_in_repeating_unit
from deepchem.utils import onek_encoding_unk
from deepchem.utils import remove_wildcard_atoms
from deepchem.utils import PolyWDGStringValidator
from rdkit import Chem


class WDGraphUtilsTest(unittest.TestCase):
    """
    Tests for polymer weighted directed graph util functions
    """

    def test_converts_smiles_without_hydrogens(self):
        smiles = 'C'
        mol = handle_hydrogen(smiles, keep_h=False, add_h=False)
        assert mol.GetNumAtoms() == 1
        assert all(atom.GetSymbol() != 'H' for atom in mol.GetAtoms())

    def test_converts_smiles_with_hydrogens(self):
        smiles = 'C'
        mol = handle_hydrogen(smiles, keep_h=False, add_h=True)
        assert mol.GetNumAtoms() == 5

    def test_converts_smiles_keeping_hydrogens(self):
        smiles = 'C([H])=O'
        mol = handle_hydrogen(smiles, keep_h=False, add_h=False)
        assert mol.GetNumAtoms() == 2
        mol = handle_hydrogen(smiles, keep_h=True, add_h=False)
        assert mol.GetNumAtoms() == 3
        mol = handle_hydrogen(smiles, keep_h=True, add_h=True)
        assert mol.GetNumAtoms() == 4

    def test_polymer_mol_with_metadata(self):
        # checking value error for incorrect number of fragments
        with self.assertRaises(ValueError):
            mol = make_polymer_mol('C.C', [1], True, False)

        mol = make_polymer_mol(
            'C.C',
            [1, 1],
            True,
            False,
        )
        for atom in mol.GetAtoms():
            _ = atom.GetDoubleProp('w_frag')

    def test_parse_invalid_polymer_rules(self):
        with self.assertRaises(ValueError):
            _ = parse_polymer_rules(["1-2:0.5"])

        with self.assertRaises(ValueError):
            _ = parse_polymer_rules(["1-2:0.3:0.3:0.3"])

        with self.assertRaises(ValueError):
            _ = parse_polymer_rules(["1-2-3:0.5:0.5"])

        with self.assertRaises(ValueError):
            _ = parse_polymer_rules(["1:1.0:0.0"])

    def test_parse_dop(self):
        _, dop = parse_polymer_rules(["1-2:0.5:0.5~10"])
        assert dop == 2

    def test_tag_atoms_in_repeating_unit(self):
        mol, _ = tag_atoms_in_repeating_unit(Chem.MolFromSmiles('[1*]CC.C[2*]'))
        assert mol.GetAtomWithIdx(0).GetBoolProp('core') is False
        assert mol.GetAtomWithIdx(1).GetBoolProp('core') is True
        assert mol.GetAtomWithIdx(0).GetProp('R') == ''
        assert mol.GetAtomWithIdx(1).GetProp('R') == '1*'
        assert mol.GetAtomWithIdx(3).GetProp('R') == '2*'

    def test_onek_encoding_unk(self):
        assert len(onek_encoding_unk(1, [0, 1, 2])) == 4
        assert onek_encoding_unk(0, [0, 1, 2]) == [1, 0, 0, 0]
        assert onek_encoding_unk(69, [0, 1, 2]) == [0, 0, 0, 1]

    def test_removing_wildcard_atoms(self):
        mol = Chem.MolFromSmiles("[*]CC[*]")
        rwmol = Chem.RWMol(mol)
        rwmol = remove_wildcard_atoms(rwmol)
        assert Chem.MolToSmiles(rwmol) == "CC"
        # not altering the atom if there is no wildcard
        mol = Chem.MolFromSmiles("CC")
        rwmol = Chem.RWMol(mol)
        rwmol = remove_wildcard_atoms(rwmol)
        assert Chem.MolToSmiles(rwmol) == "CC"


class PolyWDGStringValidateTest(unittest.TestCase):
    """
    Tests for validator class to validate polymer weighted string representation
    """

    def test_get_parsed_vals(self):
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|<1-2:0.5:0.5"
            _ = PolyWDGStringValidator.get_parsed_vals(datapoint)

        datapoint = "[1*]C.C[2*]|0.5|0.5|<1-2:0.5:0.5"
        monomer_mols, fragments, polymer_rules = PolyWDGStringValidator.get_parsed_vals(
            datapoint)
        assert monomer_mols == "[1*]C.C[2*]"
        assert fragments == ["0.5", "0.5"]
        assert polymer_rules == "<1-2:0.5:0.5"

    def test_get_polymer_rules(self):
        with self.assertRaises(ValueError):
            polymer_rules = "1-2:0.5:0.5"
            _ = PolyWDGStringValidator.get_polymer_rules(polymer_rules)

        polymer_rules = "<1-2:0.5:0.5"
        polymer_rules = PolyWDGStringValidator.get_polymer_rules(polymer_rules)
        assert polymer_rules == ["1-2:0.5:0.5"]

    def test_valid_validate_function(self):
        datapoint = "[1*]C.C[2*]|0.5|0.5|<1-2:0.5:0.5"
        assert PolyWDGStringValidator().validate(datapoint)

    def test_invalid_validate_function(self):
        # test for _validate_fragments
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|<1-2:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_wildcards
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[3*]|0.5|0.5|<1-2:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_wildcards
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<1-3:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

    def test_invalid_polymer_rules(self):
        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<1-3:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<13:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<13-:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<1-3-5:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<1-q:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)

        # test for _validate_polymer_rules
        with self.assertRaises(ValueError):
            datapoint = "[1*]C.C[2*]|0.5|0.5|<1-69:0.5:0.5"
            _ = PolyWDGStringValidator().validate(datapoint)



"""The functions in these utilities check that noncovalent interactions happen"""
import numpy as np
from collections import Counter
from deepchem.utils.fragment_utils import get_partial_charge
from deepchem.utils.rdkit_utils import compute_ring_center
from deepchem.utils.rdkit_utils import compute_ring_normal
from deepchem.utils.geometry_utils import angle_between
from deepchem.utils.geometry_utils import is_angle_within_cutoff


def is_salt_bridge(atom_i, atom_j):
    """Check if two atoms have correct charges to form a salt bridge"""
    if np.abs(2.0 -
              np.abs(get_partial_charge(atom_i) -
                     get_partial_charge(atom_j))) < 0.01:
        return True
    return False


def compute_salt_bridges(first, second, pairwise_distances, cutoff=5.0):
    """Find salt bridge contacts between two molecules.

    Parameters:
    -----------
    first: rdkit.rdchem.Mol
        Interacting molecules
    second: rdkit.rdchem.Mol
        Interacting molecules
    pairwise_distances: np.ndarray
        Array of pairwise interatomic distances between molecule atoms (Angstroms)
    cutoff: float
        Cutoff distance for contact consideration

    Returns:
    --------
    salt_bridge_contacts: list of tuples
        List of contacts. Tuple (i, j) indicates that atom i from
        first molecule interacts with atom j from second.
    """

    salt_bridge_contacts = []
    contacts = np.nonzero(pairwise_distances < cutoff)
    contacts = zip(contacts[0], contacts[1])
    for contact in contacts:
        first_atom = first.GetAtoms()[int(contact[0])]
        second_atom = second.GetAtoms()[int(contact[1])]
        if is_salt_bridge(first_atom, second_atom):
            salt_bridge_contacts.append(contact)
    return salt_bridge_contacts


def is_hydrogen_bond(frag1,
                     frag2,
                     contact,
                     hbond_distance_cutoff=4.0,
                     hbond_angle_cutoff=40.0):
    """
    Determine if a pair of atoms (contact = frag1_atom_index,
    frag2_atom_index) between two molecules represents a hydrogen
    bond. Returns a boolean result.

    Parameters
    ----------
    frag1: tuple
        Tuple of (coords, rdkit mol / MolecularFragment)
    frag2: tuple
        Tuple of (coords, rdkit mol / MolecularFragment)
    contact: Tuple
        Tuple of indices for (atom_i, atom_j) contact.
    hbond_distance_cutoff: float, optional
        Distance cutoff for hbond.
    hbond_angle_cutoff: float, optional
        Angle deviance cutoff for hbond
    """
    frag1_xyz, frag2_xyz = frag1[0], frag2[0]
    frag1_mol, frag2_mol = frag1[1], frag2[1]
    frag1_atom_xyz = frag1_xyz[int(contact[0])]
    frag2_atom_xyz = frag2_xyz[int(contact[1])]
    frag1_atom = frag1_mol.GetAtoms()[int(contact[0])]
    frag2_atom = frag2_mol.GetAtoms()[int(contact[1])]

    # Nitrogen has atomic number 7, and oxygen 8.
    if ((frag2_atom.GetAtomicNum() == 7 or frag2_atom.GetAtomicNum() == 8) and
        (frag1_atom.GetAtomicNum() == 7 or frag1_atom.GetAtomicNum() == 8)):
        hydrogens = []

        for i, atom in enumerate(frag2_mol.GetAtoms()):
            # If atom is a hydrogen
            if atom.GetAtomicNum() == 1:
                atom_xyz = frag2_xyz[i]
                dist = np.linalg.norm(atom_xyz - frag2_atom_xyz)
                # O-H distance is 0.96 A, N-H is 1.01 A. See http://www.science.uwaterloo.ca/~cchieh/cact/c120/bondel.html
                if dist < 1.3:
                    hydrogens.append(atom_xyz)

        for j, atom in enumerate(frag1_mol.GetAtoms()):
            # If atom is a hydrogen
            if atom.GetAtomicNum() == 1:
                atom_xyz = frag1_xyz[i]
                dist = np.linalg.norm(atom_xyz - frag1_atom_xyz)
                # O-H distance is 0.96 A, N-H is 1.01 A. See http://www.science.uwaterloo.ca/~cchieh/cact/c120/bondel.html
                if dist < 1.3:
                    hydrogens.append(atom_xyz)

        for hydrogen_xyz in hydrogens:
            hydrogen_to_frag2 = frag2_atom_xyz - hydrogen_xyz
            hydrogen_to_frag1 = frag1_atom_xyz - hydrogen_xyz
            return is_angle_within_cutoff(hydrogen_to_frag2, hydrogen_to_frag1,
                                          hbond_angle_cutoff)
    return False


def compute_hbonds_in_range(frag1, frag2, pairwise_distances, hbond_dist_bin,
                            hbond_angle_cutoff):
    """
    Find all pairs of (frag1_index_i, frag2_index_j) that hydrogen bond
    given a distance bin and an angle cutoff.

    Parameters
    ----------
    frag1: tuple
        Tuple of (coords, rdkit mol / MolecularFragment
    frag2: tuple
        Tuple of (coords, rdkit mol / MolecularFragment
    pairwise_distances:
        Matrix of shape `(N, M)` with pairwise distances between frag1/frag2.
    hbond_dist_bin: tuple
        Tuple of floats `(min_dist, max_dist)` in angstroms.
    hbond_angle_cutoffs: list[float]
        List of angles of deviances allowed for hbonds
    """

    contacts = np.nonzero((pairwise_distances > hbond_dist_bin[0]) &
                          (pairwise_distances < hbond_dist_bin[1]))
    contacts = zip(contacts[0], contacts[1])
    hydrogen_bond_contacts = []
    for contact in contacts:
        if is_hydrogen_bond(frag1, frag2, contact, hbond_angle_cutoff):
            hydrogen_bond_contacts.append(contact)
    return hydrogen_bond_contacts


def compute_hydrogen_bonds(frag1, frag2, pairwise_distances, hbond_dist_bins,
                           hbond_angle_cutoffs):
    """Computes hydrogen bonds between proteins and ligands.

    Returns a list of sublists. Each sublist is a series of tuples
    of (protein_index_i, ligand_index_j) that represent a hydrogen
    bond. Each sublist represents a different type of hydrogen
    bond.

    Parameters
    ----------
    frag1: tuple
        Tuple of (coords, rdkit mol / MolecularFragment
    frag2: tuple
        Tuple of (coords, rdkit mol / MolecularFragment
    pairwise_distances:
        Matrix of shape `(N, M)` with pairwise distances between frag1/frag2.
    hbond_dist_bins: list[tuple]
        List of tuples of hbond distance ranges.
    hbond_angle_cutoffs: list[float]
        List of angles of deviances allowed for hbonds

    Returns
    -------
    List
        A list of hydrogen bond contacts.
    """

    hbond_contacts = []
    for i, hbond_dist_bin in enumerate(hbond_dist_bins):
        hbond_angle_cutoff = hbond_angle_cutoffs[i]
        hbond_contacts.append(
            compute_hbonds_in_range(frag1, frag2, pairwise_distances,
                                    hbond_dist_bin, hbond_angle_cutoff))
    return (hbond_contacts)


def compute_cation_pi(mol1, mol2, charge_tolerance=0.01, **kwargs):
    """Finds aromatic rings in mo1 and cations in mol2 that interact with each other.

    Parameters:
    -----------
    mol1: rdkit.rdchem.Mol
        Molecule to look for interacting rings
    mol2: rdkit.rdchem.Mol
        Molecule to look for interacting cations
    charge_tolerance: float
        Atom is considered a cation if its formal charge is greater
        than 1 - charge_tolerance
    **kwargs:
        Arguments that are passed to is_cation_pi function

    Returns:
    --------
    mol1_pi: dict
        Dictionary that maps atom indices (from mol1) to the number of cations
        (in mol2) they interact with
    mol2_cation: dict
        Dictionary that maps atom indices (from mol2) to the number of aromatic
        atoms (in mol1) they interact with
    """
    mol1_pi = Counter()
    mol2_cation = Counter()
    conformer = mol2.GetConformer()

    aromatic_atoms = set(atom.GetIdx() for atom in mol1.GetAromaticAtoms())
    from rdkit import Chem
    rings = [list(r) for r in Chem.GetSymmSSSR(mol1)]

    for ring in rings:
        # if ring from mol1 is aromatic
        if set(ring).issubset(aromatic_atoms):
            ring_center = compute_ring_center(mol1, ring)
            ring_normal = compute_ring_normal(mol1, ring)

            for atom in mol2.GetAtoms():
                # ...and atom from mol2 is a cation
                if atom.GetFormalCharge() > 1.0 - charge_tolerance:
                    cation_position = np.array(
                        conformer.GetAtomPosition(atom.GetIdx()))
                    # if angle and distance are correct
                    if is_cation_pi(cation_position, ring_center, ring_normal,
                                    **kwargs):
                        # count atoms forming a contact
                        mol1_pi.update(ring)
                        mol2_cation.update([atom.GetIndex()])
    return mol1_pi, mol2_cation


def is_cation_pi(cation_position,
                 ring_center,
                 ring_normal,
                 dist_cutoff=6.5,
                 angle_cutoff=30.0):
    """Check if a cation and an aromatic ring form contact.

    Parameters:
    -----------
    ring_center: np.ndarray
        Positions of ring center. Can be computed with the compute_ring_center
        function.
    ring_normal: np.ndarray
        Normal of ring. Can be computed with the compute_ring_normal function.
    dist_cutoff: float
        Distance cutoff. Max allowed distance between ring center
        and cation (in Angstroms).
    angle_cutoff: float
        Angle cutoff. Max allowed deviation from the ideal (0deg)
        angle between ring normal and vector pointing from ring
        center to cation (in degrees).
    """
    cation_to_ring_vec = cation_position - ring_center
    dist = np.linalg.norm(cation_to_ring_vec)
    angle = angle_between(cation_to_ring_vec, ring_normal) * 180. / np.pi
    if ((angle < angle_cutoff or angle > 180.0 - angle_cutoff) and
        (dist < dist_cutoff)):
        return True
    return False


def compute_pi_stack(mol1,
                     mol2,
                     pairwise_distances=None,
                     dist_cutoff=4.4,
                     angle_cutoff=30.):
    """Find aromatic rings in both molecules that form pi-pi contacts.
    For each atom in the contact, count number of atoms in the other molecule
    that form this contact.

    Pseudocode:

    for each aromatic ring in mol1:
        for each aromatic ring in mol2:
            compute distance between centers
            compute angle between normals
        if it counts as parallel pi-pi:
            count interacting atoms
        if it counts as pi-T:
            count interacting atoms

    Parameters:
    -----------
    mol1: rdkit.rdchem.Mol
        First molecule.
    mol2: rdkit.rdchem.Mol
        Second molecule.
    pairwise_distances: np.ndarray (optional)
        Array of pairwise interatomic distances (Angstroms)
    dist_cutoff: float
        Distance cutoff. Max allowed distance between the ring center (Angstroms).
    angle_cutoff: float
        Angle cutoff. Max allowed deviation from the ideal angle between rings.

    Returns:
    --------
    mol1_pi_t, mol1_pi_parallel, mol2_pi_t, mol2_pi_parallel: dict
        Dictionaries mapping atom indices to number of atoms they interact with.
        Separate dictionary is created for each type of pi stacking (parallel and
        T-shaped) and each molecule (mol1 and mol2).
    """

    mol1_pi_parallel = Counter()
    mol1_pi_t = Counter()
    mol2_pi_parallel = Counter()
    mol2_pi_t = Counter()

    mol1_aromatic_rings = []
    mol2_aromatic_rings = []
    from rdkit import Chem
    for mol, ring_list in ((mol1, mol1_aromatic_rings), (mol2,
                                                         mol2_aromatic_rings)):
        aromatic_atoms = {atom.GetIdx() for atom in mol.GetAromaticAtoms()}
        for ring in Chem.GetSymmSSSR(mol):
            # if ring is aromatic
            if set(ring).issubset(aromatic_atoms):
                # save its indices, center, and normal
                ring_center = compute_ring_center(mol, ring)
                ring_normal = compute_ring_normal(mol, ring)
                ring_list.append((ring, ring_center, ring_normal))

    # remember mol1-mol2 pairs we already counted
    counted_pairs_parallel = set()
    counted_pairs_t = set()
    for prot_ring, prot_ring_center, prot_ring_normal in mol1_aromatic_rings:
        for lig_ring, lig_ring_center, lig_ring_normal in mol2_aromatic_rings:
            if is_pi_parallel(prot_ring_center,
                              prot_ring_normal,
                              lig_ring_center,
                              lig_ring_normal,
                              angle_cutoff=angle_cutoff,
                              dist_cutoff=dist_cutoff):
                prot_to_update = set()
                lig_to_update = set()
                for prot_atom_idx in prot_ring:
                    for lig_atom_idx in lig_ring:
                        if (prot_atom_idx,
                                lig_atom_idx) not in counted_pairs_parallel:
                            # if this pair is new, count atoms forming a contact
                            prot_to_update.add(prot_atom_idx)
                            lig_to_update.add(lig_atom_idx)
                            counted_pairs_parallel.add(
                                (prot_atom_idx, lig_atom_idx))

                mol1_pi_parallel.update(prot_to_update)
                mol2_pi_parallel.update(lig_to_update)

            if is_pi_t(prot_ring_center,
                       prot_ring_normal,
                       lig_ring_center,
                       lig_ring_normal,
                       angle_cutoff=angle_cutoff,
                       dist_cutoff=dist_cutoff):
                prot_to_update = set()
                lig_to_update = set()
                for prot_atom_idx in prot_ring:
                    for lig_atom_idx in lig_ring:
                        if (prot_atom_idx, lig_atom_idx) not in counted_pairs_t:
                            # if this pair is new, count atoms forming a contact
                            prot_to_update.add(prot_atom_idx)
                            lig_to_update.add(lig_atom_idx)
                            counted_pairs_t.add((prot_atom_idx, lig_atom_idx))

                mol1_pi_t.update(prot_to_update)
                mol2_pi_t.update(lig_to_update)

    return (mol1_pi_t, mol1_pi_parallel, mol2_pi_t, mol2_pi_parallel)


def is_pi_t(ring1_center,
            ring1_normal,
            ring2_center,
            ring2_normal,
            dist_cutoff=5.5,
            angle_cutoff=30.0):
    """Check if two aromatic rings form a T-shaped pi-pi contact.

    Parameters:
    -----------
    ring1_center, ring2_center: np.ndarray
        Positions of centers of the two rings. Can be computed with the
        compute_ring_center function.
    ring1_normal, ring2_normal: np.ndarray
        Normals of the two rings. Can be computed with the compute_ring_normal
        function.
    dist_cutoff: float
        Distance cutoff. Max allowed distance between the ring center (Angstroms).
    angle_cutoff: float
        Angle cutoff. Max allowed deviation from the ideal (90deg) angle between
        the rings (in degrees).
    """
    dist = np.linalg.norm(ring1_center - ring2_center)
    angle = angle_between(ring1_normal, ring2_normal) * 180 / np.pi
    if ((90.0 - angle_cutoff < angle < 90.0 + angle_cutoff) and
            dist < dist_cutoff):
        return True
    return False


def is_pi_parallel(ring1_center: np.ndarray,
                   ring1_normal: np.ndarray,
                   ring2_center: np.ndarray,
                   ring2_normal: np.ndarray,
                   dist_cutoff: float = 8.0,
                   angle_cutoff: float = 30.0) -> bool:
    """Check if two aromatic rings form a parallel pi-pi contact.

    Parameters
    ----------
    ring1_center, ring2_center: np.ndarray
        Positions of centers of the two rings. Can be computed with the
        compute_ring_center function.
    ring1_normal, ring2_normal: np.ndarray
        Normals of the two rings. Can be computed with the compute_ring_normal
        function.
    dist_cutoff: float
        Distance cutoff. Max allowed distance between the ring center (Angstroms).
    angle_cutoff: float
        Angle cutoff. Max allowed deviation from the ideal (0deg) angle between
        the rings (in degrees).

    Returns
    -------
    bool
        True if two aromatic rings form a parallel pi-pi.
    """

    dist = np.linalg.norm(ring1_center - ring2_center)
    angle = angle_between(ring1_normal, ring2_normal) * 180 / np.pi
    if ((angle < angle_cutoff or angle > 180.0 - angle_cutoff) and
            dist < dist_cutoff):
        return True
    return False


def compute_binding_pocket_cation_pi(mol1, mol2, **kwargs):
    """Finds cation-pi interactions between mol1 and mol2.

    Parameters:
    -----------
    mol1: rdkit.rdchem.Mol
        Interacting molecules
    mol2: rdkit.rdchem.Mol
        Interacting molecules
    **kwargs:
        Arguments that are passed to compute_cation_pi function

    Returns:
    --------
    mol1_cation_pi, mol2_cation_pi: dict
        Dictionaries that maps atom indices to the number of cations/aromatic
        atoms they interact with
    """
    # find interacting rings from mol1 and cations from mol2
    mol1_pi, mol2_cation = compute_cation_pi(mol1, mol2, **kwargs)
    # find interacting cations from mol1 and rings from mol2
    mol2_pi, mol1_cation = compute_cation_pi(mol2, mol1, **kwargs)

    # merge counters
    mol1_cation_pi = Counter()
    mol1_cation_pi.update(mol1_pi)
    mol1_cation_pi.update(mol1_cation)

    mol2_cation_pi = Counter()
    mol2_cation_pi.update(mol2_pi)
    mol2_cation_pi.update(mol2_cation)

    return mol1_cation_pi, mol2_cation_pi



import torch
from collections import defaultdict
from deepchem.utils import get_option
from typing import Union, List, Optional, Mapping, Callable, Dict
from deepchem.utils.periodic_table_utils import atom_bragg_radii, atom_expected_radii, get_period
from deepchem.utils.dft_utils import DasguptaTrunc, NWChemTrunc, NoTrunc, Lattice, BaseGrid, RadialGrid, LogM3Transformation, TreutlerM4Transformation, DE2Transformation, LebedevGrid, TruncatedLebedevGrid, BeckeGrid, PBCBeckeGrid

# list of alphas for de2 transformation from https://onlinelibrary.wiley.com/doi/epdf/10.1002/jcc.24761
__sg2_dasgupta_alphas = defaultdict(
    lambda: 1.0, {
        1: 2.6,
        3: 3.2,
        4: 2.4,
        5: 2.4,
        6: 2.2,
        7: 2.2,
        8: 2.2,
        9: 2.2,
        11: 3.2,
        12: 2.4,
        13: 2.5,
        14: 2.3,
        15: 2.5,
        16: 2.5,
        17: 2.5,
    })
__sg3_dasgupta_alphas = defaultdict(
    lambda: 1.0, {
        1: 2.7,
        3: 3.0,
        4: 2.4,
        5: 2.4,
        6: 2.4,
        7: 2.4,
        8: 2.6,
        9: 2.1,
        11: 3.2,
        12: 2.6,
        13: 2.6,
        14: 2.8,
        15: 2.4,
        16: 2.4,
        17: 2.6,
    })

# list of optimized xi for M4 transformation from Treutler Table I
# https://doi.org/10.1063/1.469408
__treutler_xi = defaultdict(
    lambda: 1.0, {
        1: 0.8,
        2: 0.9,
        3: 1.8,
        4: 1.4,
        5: 1.3,
        6: 1.1,
        7: 0.9,
        8: 0.9,
        9: 0.9,
        10: 0.9,
        11: 1.4,
        12: 1.3,
        13: 1.3,
        14: 1.2,
        15: 1.1,
        16: 1.0,
        17: 1.0,
        18: 1.0,
        19: 1.5,
        20: 1.4,
        21: 1.3,
        22: 1.2,
        23: 1.2,
        24: 1.2,
        25: 1.2,
        26: 1.2,
        27: 1.2,
        28: 1.1,
        29: 1.1,
        30: 1.1,
        31: 1.1,
        32: 1.0,
        33: 0.9,
        34: 0.9,
        35: 0.9,
        36: 0.9,
    })

# number of angular points to precision
__nang2prec = {
    6: 3,
    14: 5,
    26: 7,
    38: 9,
    50: 11,
    74: 13,
    86: 15,
    110: 17,
    146: 19,
    170: 21,
    194: 23,
    230: 25,
    266: 27,
    302: 29,
    350: 31,
    434: 35,
    590: 41,
    770: 47,
    974: 53,
    1202: 59,
    1454: 65,
    1730: 71,
    2030: 77,
    2354: 83,
    2702: 89,
    3074: 95,
    3470: 101,
    3890: 107,
    4334: 113,
    4802: 119,
    5294: 125,
    5810: 131,
}

_dtype = torch.double
_device = torch.device("cpu")


def get_grid(atomzs: Union[List[int], torch.Tensor],
             atompos: torch.Tensor,
             *,
             lattice: Optional[Lattice] = None,
             nr: Union[int, Callable[[int], int]] = 99,
             nang: Union[int, Callable[[int], int]] = 590,
             radgrid_generator: str = "uniform",
             radgrid_transform: str = "sg2-dasgupta",
             atom_radii: str = "expected",
             multiatoms_scheme: str = "becke",
             truncate: Optional[str] = "dasgupta",
             dtype: torch.dtype = _dtype,
             device: torch.device = _device) -> BaseGrid:
    """Get the Molecular Grid.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import get_grid
    >>> grid = get_grid(torch.tensor([1]), torch.tensor([[0,0,0]], dtype=torch.float64))
    >>> grid.get_rgrid().shape
    torch.Size([16710, 3])

    Parameters
    ----------
    atomzs: Union[List[int], torch.Tensor]
        Atomic Number of the atoms in the molecule.
    atompos: torch.Tensor
        Positions of the atoms in 3d. (natoms, ndim)
    lattice: Optional[Lattice] (Default None)
        Lattice Object descibing periodicity.
    nr: Union[int, Callable[[int], int]] (default 99)
        Number of radial points.
    nang: Union[int, Callable[[int], int]] = 590
        Number of angular points.
    radgrid_generator: str (default "uniform")
        Radial grid generator.
    radgrid_transform: str (default "sg2-dasgupta")
        Radial grid transformation.
    atom_radii: str (default "expected")
        Atom radii to use.
    multiatoms_scheme: str (default "becke")
        Multiatoms scheme to use.
    truncate: Optional[str] (default "dasgupta")
        Truncation rule to use.
    dtype: torch.dtype (default _dtype)
        Data type to use.
    device: torch.device (default _device)
        Device to use.

    Returns
    -------
    BaseGrid
        Molecular Grid.

    """
    # atompos: (natoms, ndim)
    assert atompos.ndim == 2
    assert atompos.shape[-2] == len(atomzs)

    # convert the atomzs to a list of integers
    if isinstance(atomzs, torch.Tensor):
        assert atomzs.ndim == 1
        atomzs_list = [a.item() for a in atomzs]
    else:
        atomzs_list = list(atomzs)

    # get the atom radii list
    atom_radii_options: Mapping[str, Union[List[float]]] = {
        "expected": atom_expected_radii,
        "bragg": atom_bragg_radii,
    }
    atom_radii_list = get_option("atom radii", atom_radii, atom_radii_options)
    atomradii = torch.tensor([atom_radii_list[atomz] for atomz in atomzs_list],
                             dtype=dtype,
                             device=device)

    # construct the radial grid transformation as a function of atom z
    radgrid_tf_options = {
        "sg2-dasgupta":
            lambda atz: DE2Transformation(alpha=__sg2_dasgupta_alphas[atz],
                                          rmin=1e-7,
                                          rmax=15 * atom_radii_list[atz]),
        "sg3-dasgupta":
            lambda atz: DE2Transformation(alpha=__sg3_dasgupta_alphas[atz],
                                          rmin=1e-7,
                                          rmax=15 * atom_radii_list[atz]),
        "logm3":
            lambda atz: LogM3Transformation(ra=atom_radii_list[atz]),
        "treutlerm4":
            lambda atz: TreutlerM4Transformation(xi=__treutler_xi[atz],
                                                 alpha=0.6),
    }
    radgrid_tf = get_option("radial grid transformation", radgrid_transform,
                            radgrid_tf_options)

    # get the precisions
    if isinstance(nang, int):
        prec: Union[int, Callable[[int],
                                  int]] = get_option("number of angular points",
                                                     nang, __nang2prec)
    else:

        def _prec_fcn(atz: int) -> int:
            assert callable(nang)
            return get_option("number of angular points", nang(atz),
                              __nang2prec)

        prec = _prec_fcn

    # wrap up a function to get the nr
    def _get_nr(nr: Union[int, Callable[[int], int]], atz: int) -> int:
        """Get the number of radial points for the atom z.

        Parameters
        ----------
        nr: Union[int, Callable[[int], int]]
            Number of radial points.
        atz: int
            Atomic Number of the atom.

        Returns
        -------
        int
            Number of radial points.

        """
        if isinstance(nr, int):
            return nr
        else:
            return nr(atz)

    # get the truncation rule as a function to avoid unnecessary evaluation
    trunc_options = {
        "dasgupta":
            lambda: DasguptaTrunc(nr),
        "nwchem":
            lambda: NWChemTrunc(atom_radii_list,
                                prec,
                                list(__nang2prec.values()),
                                dtype=dtype,
                                device=device),
        "no":
            lambda: NoTrunc(),
    }
    truncate_str = truncate if truncate is not None else "no"
    trunc = get_option("truncation rule", truncate_str, trunc_options)()

    sphgrids: List[BaseGrid] = []
    sphgrids_dict: Dict[int, BaseGrid] = {}
    for atz in atomzs_list:
        if atz in sphgrids_dict:
            sphgrids.append(sphgrids_dict[atz])
            continue

        nr_value = _get_nr(nr, atz)
        radgrid = RadialGrid(nr_value,
                             grid_integrator=radgrid_generator,
                             grid_transform=radgrid_tf(atz),
                             dtype=dtype,
                             device=device)
        if trunc.to_truncate(atz):
            rad_slices = trunc.rad_slices(atz, radgrid)
            radgrids: List[BaseGrid] = [radgrid[sl] for sl in rad_slices]
            precs = trunc.precs(atz, radgrid)
            sphgrid = TruncatedLebedevGrid(radgrids, precs)  # type: ignore
        else:
            sphgrid = LebedevGrid(
                radgrid,
                prec=_get_nr(
                    prec,  # type: ignore
                    atz))
        sphgrids_dict[atz] = sphgrid
        sphgrids.append(sphgrid)

    # get the multi atoms grid
    # the values are a function to avoid constructing it unnecessarily
    if lattice is None:
        multiatoms_options: Mapping[str, Callable[[], BaseGrid]] = {
            "becke":
                lambda: BeckeGrid(
                    sphgrids,  # type: ignore
                    atompos,
                    atomradii=atomradii),
            "treutler":
                lambda: BeckeGrid(
                    sphgrids,  # type: ignore
                    atompos,
                    atomradii=atomradii,
                    ratom_adjust="treutler"),
        }
    else:
        assert isinstance(lattice, Lattice)
        multiatoms_options = {
            "becke":
                lambda: PBCBeckeGrid(
                    sphgrids,  # type: ignore
                    atompos,
                    lattice=lattice),
            "treutler":
                lambda: PBCBeckeGrid(
                    sphgrids,  # type: ignore
                    atompos,
                    lattice=lattice,
                    ratom_adjust="treutler"),
        }
    grid = get_option("multiatoms scheme", multiatoms_scheme,
                      multiatoms_options)()
    return grid


def get_predefined_grid(grid_inp: Union[int, str],
                        atomzs: Union[List[int], torch.Tensor],
                        atompos: torch.Tensor,
                        *,
                        lattice: Optional[Lattice] = None,
                        dtype: torch.dtype = _dtype,
                        device: torch.device = _device) -> BaseGrid:
    """Returns the predefined grid object given the grid name.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import get_predefined_grid
    >>> grid = get_predefined_grid(3, torch.tensor([2]), torch.tensor([[0,0,0]], dtype=torch.float64))
    >>> grid.get_rgrid().shape
    torch.Size([8608, 3])

    Parameters
    ----------
    grid_inp: Union[int, str]
        Grid name or number of angular points.
    atomzs: Union[List[int], torch.Tensor]
        Atomic Number of the atoms in the molecule.
    atompos: torch.Tensor
        Positions of the atoms in 3d. (natoms, ndim)
    lattice: Optional[Lattice] (Default None)
        Lattice Object descibing periodicity.
    dtype: torch.dtype (default _dtype)
        Data type to use.
    device: torch.device (default _device)
        Device to use.

    Returns
    -------
    BaseGrid
        Molecular Grid.

    """
    if isinstance(grid_inp, str):
        if grid_inp == "sg2":
            return get_grid(
                atomzs,
                atompos,
                lattice=lattice,
                nr=75,
                nang=302,
                radgrid_generator="uniform",
                radgrid_transform="sg2-dasgupta",
                # using expected from de2 ref: DOI 10.1007/s00214-012-1169-z
                atom_radii="expected",
                multiatoms_scheme="becke",
                truncate="dasgupta",
                dtype=dtype,
                device=device)
        elif grid_inp == "sg3":
            return get_grid(
                atomzs,
                atompos,
                lattice=lattice,
                nr=99,
                nang=590,
                radgrid_generator="uniform",
                radgrid_transform="sg3-dasgupta",
                # using expected from de2 ref: DOI 10.1007/s00214-012-1169-z
                atom_radii="expected",
                multiatoms_scheme="becke",
                truncate="dasgupta",
                dtype=dtype,
                device=device)
        else:
            raise ValueError(f"Unknown grid name: {grid_inp}")
    elif isinstance(grid_inp, int):
        # the numbers are taken from pyscf
        # https://github.com/pyscf/pyscf/blob/21a0b8d19ec58d8e70901c77783fa72ac47c1745/pyscf/dft/gen_grid.py#L596
        # period:     1,   2,   3,   4,   5,   6,   7    # level
        nr_list = (
            (10, 15, 20, 30, 35, 40, 50),  # 0
            (30, 40, 50, 60, 65, 70, 75),  # 1
            (40, 60, 65, 75, 80, 85, 90),  # 2
            (50, 75, 80, 90, 95, 100, 105),  # 3
            (60, 90, 95, 105, 110, 115, 120),  # 4
            (70, 105, 110, 120, 125, 130, 135),  # 5
            (80, 120, 125, 135, 140, 145, 150),  # 6
            (90, 135, 140, 150, 155, 160, 165),  # 7
            (100, 150, 155, 165, 170, 175, 180),  # 8
            (200, 200, 200, 200, 200, 200, 200),  # 9
        )

        # period:        1,    2,    3,    4,    5,    6,    7    # level
        nang_list = [
            [50, 86, 110, 110, 110, 110, 110],  # 0
            [110, 194, 194, 194, 194, 194, 194],  # 1
            [194, 302, 302, 302, 302, 302, 302],  # 2
            [302, 302, 434, 434, 434, 434, 434],  # 3
            [434, 590, 590, 590, 590, 590, 590],  # 4
            [590, 770, 770, 770, 770, 770, 770],  # 5
            [770, 974, 974, 974, 974, 974, 974],  # 6
            [974, 1202, 1202, 1202, 1202, 1202, 1202],  # 7
            [1202, 1202, 1202, 1202, 1202, 1202, 1202],  # 8
            [1454, 1454, 1454, 1454, 1454, 1454, 1454],  # 9
        ]

        nr_list2 = nr_list[grid_inp]
        nang_list2 = nang_list[grid_inp]

        def get_nr(atz: int) -> int:
            period = get_period(atz)
            return nr_list2[period - 1]

        def get_nang(atz: int) -> int:
            period = get_period(atz)
            return nang_list2[period - 1]

        return get_grid(atomzs,
                        atompos,
                        lattice=lattice,
                        nr=get_nr,
                        nang=get_nang,
                        radgrid_generator="chebyshev2",
                        radgrid_transform="treutlerm4",
                        atom_radii="bragg",
                        multiatoms_scheme="treutler",
                        truncate="nwchem",
                        dtype=dtype,
                        device=device)
    else:
        raise TypeError("Unknown type of grid_inp: %s" % type(grid_inp))



from typing import Callable, Mapping, Any, Sequence, Union
import torch
from deepchem.utils.differentiation_utils.optimize.rootsolver import broyden1, broyden2, \
    linearmixing
from deepchem.utils.differentiation_utils.optimize.equilibrium import anderson_acc
from deepchem.utils.differentiation_utils.optimize.minimizer import gd, adam
from deepchem.utils.differentiation_utils.solve import solve
from deepchem.utils.differentiation_utils.grad import jac
from deepchem.utils.differentiation_utils import get_pure_function, make_sibling
from deepchem.utils.differentiation_utils.misc import get_method
from deepchem.utils import TensorNonTensorSeparator

__all__ = ["equilibrium", "rootfinder", "minimize"]

_RF_METHODS = {
    "broyden1": broyden1,
    "broyden2": broyden2,
    "linearmixing": linearmixing,
}

_EQUIL_METHODS = {
    # equilibrium can use all rootfinder methods, but there are several methods developed specifically for
    # equilibrium (or fixed-point iterations). This dictionary gives the list of those special methods.
    "anderson_acc": anderson_acc,
}

_OPT_METHODS = {
    "gd": gd,
    "adam": adam,
}


def rootfinder(fcn: Callable[..., torch.Tensor],
               y0: torch.Tensor,
               params: Sequence[Any] = [],
               bck_options: Mapping[str, Any] = {},
               method: Union[str, Callable, None] = None,
               **fwd_options) -> torch.Tensor:
    r"""
    Solving the rootfinder equation of a given function,

    .. math::

        \mathbf{0} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the :math:`\mathbf{0}` as the output.

    Parameters
    ----------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or callable or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{0} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    >>> import torch
    >>> def func1(y, A):  # example function
    ...     return torch.tanh(A @ y + 0.1) + y / 2.0
    >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
    >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
    >>> yroot = rootfinder(func1, y0, params=(A,))
    >>> print(yroot)
    tensor([[-0.0459],
            [-0.0663]], grad_fn=<_RootFinderBackward>)

    """

    pfunc = get_pure_function(fcn)
    fwd_options["method"] = _get_rootfinder_default_method(method)
    return _RootFinder.apply(pfunc, y0, pfunc, "rootfinder", fwd_options,
                             bck_options, len(params), *params,
                             *pfunc.objparams())


def equilibrium(fcn: Callable[..., torch.Tensor],
                y0: torch.Tensor,
                params: Sequence[Any] = [],
                bck_options: Mapping[str, Any] = {},
                method: Union[str, Callable, None] = None,
                **fwd_options) -> torch.Tensor:
    r"""
    Solving the equilibrium equation of a given function,

    .. math::

        \mathbf{y} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the same :math:`\mathbf{y}` as the output.

    Parameters
    ----------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{y} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    >>> import torch
    >>> def func1(y, A):  # example function
    ...     return torch.tanh(A @ y + 0.1) + y / 2.0
    >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
    >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
    >>> yequil = equilibrium(func1, y0, params=(A,))
    >>> print(yequil)
    tensor([[ 0.2313],
            [-0.5957]], grad_fn=<_RootFinderBackward>)

    Note
    ----
    * This is a direct implementation of finding the root of
      :math:`\mathbf{g}(\mathbf{y}, \theta) = \mathbf{y} - \mathbf{f}(\mathbf{y}, \theta)`

    """
    pfunc = get_pure_function(fcn)

    @make_sibling(pfunc)
    def new_fcn(y, *params):
        return y - pfunc(y, *params)

    method = _get_equilibrium_default_method(method)
    fwd_options["method"] = method
    fwd_fcn = pfunc if method in _EQUIL_METHODS else new_fcn
    alg_type = "equilibrium" if method in _EQUIL_METHODS else "rootfinder"
    return _RootFinder.apply(new_fcn, y0, fwd_fcn, alg_type, fwd_options,
                             bck_options, len(params), *params,
                             *pfunc.objparams())


def minimize(fcn: Callable[..., torch.Tensor],
             y0: torch.Tensor,
             params: Sequence[Any] = [],
             bck_options: Mapping[str, Any] = {},
             method: Union[None, str, Callable] = None,
             **fwd_options) -> torch.Tensor:
    r"""
    Solve the unbounded minimization problem:

    .. math::

        \mathbf{y^*} = \arg\min_\mathbf{y} f(\mathbf{y}, \theta)

    to find the best :math:`\mathbf{y}` that minimizes the output of the
    function :math:`f`.

    Parameters
    ----------
    fcn: callable
        The function to be optimized with output tensor with 1 element.
    y0: torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params: list
        Sequence of any other parameters to be put in ``fcn``
    bck_options: dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method: str or callable or None
        Minimization method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution of the minimization with shape ``(*ny)``

    Example
    -------
    >>> import torch
    >>> def func1(y, A):  # example function
    ...     return torch.sum((A @ y)**2 + y / 2.0)
    >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
    >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
    >>> ymin = minimize(func1, y0, params=(A,))
    >>> print(ymin)
    tensor([[-0.0519],
            [-0.2684]], grad_fn=<_RootFinderBackward>)

    """

    assert not torch.is_complex(y0), \
        "complex number is not supported on xitorch.optimize.rootfinder at the moment"

    pfunc = get_pure_function(fcn)

    fwd_options["method"] = _get_minimizer_default_method(method)
    method = fwd_options["method"]

    # minimization can use rootfinder algorithm, so check if it is actually
    # using the optimization algorithm, not the rootfinder algorithm
    opt_method = method not in _RF_METHODS.keys()

    # the rootfinder algorithms are designed to move to the opposite direction
    # of the output of the function, so the output of this function is just
    # the grad of z w.r.t. y
    # if it is going to optimization method, then also returns the value

    @make_sibling(pfunc)
    def _min_fwd_fcn(y, *params):
        with torch.enable_grad():
            y1 = y.clone().requires_grad_()
            z = pfunc(y1, *params)
        grady, = torch.autograd.grad(z, (y1,),
                                     retain_graph=True,
                                     create_graph=torch.is_grad_enabled())
        return z, grady

    @make_sibling(_min_fwd_fcn)
    def _rf_fcn(y, *params):
        z, grady = _min_fwd_fcn(y, *params)
        return grady

    # if using the optimization algorithm, then the forward function is the one
    # that returns f and grad
    if opt_method:
        _fwd_fcn = _min_fwd_fcn
    # if it is just using the rootfinder algorithm, then the forward function
    # is the one that returns only the grad
    else:
        _fwd_fcn = _rf_fcn

    alg_type = "minimizer" if opt_method else "rootfinder"
    return _RootFinder.apply(_rf_fcn, y0, _fwd_fcn, alg_type, fwd_options,
                             bck_options, len(params), *params,
                             *pfunc.objparams())


class _RootFinder(torch.autograd.Function):

    @staticmethod
    def forward(ctx, fcn, y0, fwd_fcn, alg_type, options, bck_options, nparams,
                *allparams):
        """Forward method for the rootfinder, minimizer, and equilibrium

        Parameters
        ----------
        fcn:
            a function that returns what has to be 0 (will be used in the
            backward, not used in the forward). For minimization, it is
            the gradient
        y0:
            initial guess
        fwd_fcn:
            a function that will be executed in the forward method
            (unused in the backward)
        alg_type:
            the type of algorithm: "rootfinder", "minimizer", or "equilibrium"
        options:
            options for the forward method
        bck_options:
            options for the backward method
        nparams:
            number of parameters
        allparams:
            all parameters (including the non-tensor parameters)

        This class is also used for minimization, where fcn and fwd_fcn might
        be slightly different

        Returns
        -------
        torch.tensor
            The solution of the rootfinder, minimizer, or equilibrium

        """

        # set default options
        config = options
        ctx.bck_options = bck_options

        params = allparams[:nparams]
        objparams = allparams[nparams:]

        with fwd_fcn.useobjparams(objparams):

            method = config.pop("method")
            methods = {
                "minimizer": _OPT_METHODS,
                "rootfinder": _RF_METHODS,
                "equilibrium": _EQUIL_METHODS
            }[alg_type]
            name = alg_type
            method_fcn = get_method(name, methods, method)
            y = method_fcn(fwd_fcn, y0, params, **config)

        ctx.fcn = fcn

        # split tensors and non-tensors params
        ctx.nparams = nparams
        ctx.param_sep = TensorNonTensorSeparator(allparams)
        tensor_params = ctx.param_sep.get_tensor_params()
        ctx.save_for_backward(y, *tensor_params)

        return y

    @staticmethod
    def backward(ctx, grad_yout):
        """Backward method for the rootfinder, minimizer, and equilibrium

        Parameters
        ----------
        grad_yout: torch.tensor
            the gradient of the output of the rootfinder, minimizer, or equilibrium

        Returns
        -------
        tuple
            The gradients of the parameters

        """

        param_sep = ctx.param_sep
        yout = ctx.saved_tensors[0]
        nparams = ctx.nparams
        fcn = ctx.fcn

        # merge the tensor and nontensor parameters
        tensor_params = ctx.saved_tensors[1:]
        allparams = param_sep.reconstruct_params(tensor_params)
        params = allparams[:nparams]
        objparams = allparams[nparams:]

        # dL/df
        with ctx.fcn.useobjparams(objparams):

            jac_dfdy = jac(fcn, params=(yout, *params), idxs=[0])[0]
            gyfcn = solve(A=jac_dfdy.H,
                          B=-grad_yout.reshape(-1, 1),
                          bck_options=ctx.bck_options,
                          **ctx.bck_options)
            gyfcn = gyfcn.reshape(grad_yout.shape)

            # get the grad for the params
            with torch.enable_grad():
                tensor_params_copy = [
                    p.clone().requires_grad_() for p in tensor_params
                ]
                allparams_copy = param_sep.reconstruct_params(
                    tensor_params_copy)
                params_copy = allparams_copy[:nparams]
                objparams_copy = allparams_copy[nparams:]
                with ctx.fcn.useobjparams(objparams_copy):
                    yfcn = fcn(yout, *params_copy)

            grad_tensor_params = torch.autograd.grad(
                yfcn,
                tensor_params_copy,
                grad_outputs=gyfcn,
                create_graph=torch.is_grad_enabled(),
                allow_unused=True)
            grad_nontensor_params = [
                None for _ in range(param_sep.nnontensors())
            ]
            grad_params = param_sep.reconstruct_params(grad_tensor_params,
                                                       grad_nontensor_params)

        return (None, None, None, None, None, None, None, *grad_params)


def _get_rootfinder_default_method(
        method: Union[str, Callable,
                      None] = None) -> Union[str, Callable, None]:
    """Get the default method for the rootfinder, minimizer, and equilibrium

    Examples
    --------
    >>> _get_rootfinder_default_method(None)
    'broyden1'

    Parameters
    ----------
    method: str or None
        The method name

    Returns
    -------
    str
        The method name

    """
    if method is None:
        return "broyden1"
    else:
        return method


def _get_equilibrium_default_method(
        method: Union[str, Callable,
                      None] = None) -> Union[str, Callable, None]:
    """Get the default method for the equilibrium

    Examples
    --------
    >>> _get_equilibrium_default_method(None)
    'broyden1'

    Parameters
    ----------
    method: str or None
        The method name

    Returns
    -------
    str
        The method name

    """
    if method is None:
        return _get_rootfinder_default_method(method)
    else:
        return method


def _get_minimizer_default_method(
        method: Union[str, Callable,
                      None] = None) -> Union[str, Callable, None]:
    """Get the default method for the minimizer

    Examples
    --------
    >>> _get_minimizer_default_method(None)
    'broyden1'

    Parameters
    ----------
    method: str or None
        The method name

    Returns
    -------
    str
        The method name

    """
    if method is None:
        return "broyden1"
    else:
        return method



import re
import torch
import ctypes
import numpy as np
from typing import Tuple, Optional
from deepchem.utils import estimate_ovlp_rcut
from deepchem.utils.dft_utils.hamilton.intor.molintor import _gather_at_dims
from deepchem.utils.dft_utils.hamilton.intor.utils import np2ctypes, int2ctypes, NDIM, CGTO
from deepchem.utils.dft_utils import LibcintWrapper, get_default_kpts, get_default_options, PBCIntOption

BLKSIZE = 128  # same as lib/gto/grid_ao_drv.c


# evaluation of the gaussian basis
def evl(shortname: str,
        wrapper: LibcintWrapper,
        rgrid: torch.Tensor,
        *,
        to_transpose: bool = False) -> torch.Tensor:
    """Evaluates the Gaussian Basis

    Examples
    --------
    >>> from deepchem.utils.dft_utils import evl, AtomCGTOBasis, LibcintWrapper, loadbasis, RadialGrid
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = evl("", wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([6, 100])

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate
    to_transpose: book (default False)
        True for transposing the matrix.

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto
    """
    # expand ao_to_atom to have shape of (nao, ndim)
    ao_to_atom = wrapper.ao_to_atom().unsqueeze(-1).expand(-1, NDIM)

    # rgrid: (ngrid, ndim)
    return _EvalGTO.apply(
        # tensors
        *wrapper.params,
        rgrid,

        # nontensors or int tensors
        ao_to_atom,
        wrapper,
        shortname,
        to_transpose)


def pbc_evl(shortname: str,
            wrapper: LibcintWrapper,
            rgrid: torch.Tensor,
            kpts: Optional[torch.Tensor] = None,
            options: Optional[PBCIntOption] = None) -> torch.Tensor:
    """evaluate the basis in periodic boundary condition,
    i.e. evaluate sum_L exp(i*k*L) * phi(r - L)

    Examples
    --------
    >>> from deepchem.utils.dft_utils import pbc_evl, AtomCGTOBasis, LibcintWrapper, loadbasis, Lattice, RadialGrid
    >>> dtype = torch.float64
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=dtype)
    >>> lattice = Lattice(a)
    >>> wrap = LibcintWrapper(atombases, True, lattice)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = pbc_evl("", wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([1, 6, 100])

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate. (ngrid, ndim)
    kpts: Optional[torch.Tensor] (default None)
        k-points in the Hamiltonian. (nkpts, ndim)
    options: Optional[PBCIntOption] (default None)

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto. (*ncomp, nkpts, nao, ngrid)

    """
    # get the default arguments
    kpts1 = get_default_kpts(kpts, dtype=wrapper.dtype, device=wrapper.device)
    options1 = get_default_options(options)

    # get the shifts
    coeffs, alphas, _ = wrapper.params
    rcut = estimate_ovlp_rcut(options1.precision, coeffs, alphas)
    assert wrapper.lattice is not None
    ls = wrapper.lattice.get_lattice_ls(rcut=rcut)  # (nls, ndim)

    # evaluate the gto
    exp_ikl = torch.exp(
        1j * torch.matmul(kpts1, ls.transpose(-2, -1)))  # (nkpts, nls)
    rgrid_shift = rgrid - ls.unsqueeze(-2)  # (nls, ngrid, ndim)
    ao = evl(shortname, wrapper,
             rgrid_shift.reshape(-1, NDIM))  # (*ncomp, nao, nls * ngrid)
    ao = ao.reshape(*ao.shape[:-1], ls.shape[0],
                    -1)  # (*ncomp, nao, nls, ngrid)
    out = torch.einsum("kl,...alg->...kag", exp_ikl,
                       ao.to(exp_ikl.dtype))  # (*ncomp, nkpts, nao, ngrid)
    return out


# shortcuts
def eval_gto(wrapper: LibcintWrapper,
             rgrid: torch.Tensor,
             *,
             to_transpose: bool = False) -> torch.Tensor:
    """Evaluates GTO

    Examples
    --------
    >>> from deepchem.utils.dft_utils import eval_gto, AtomCGTOBasis, LibcintWrapper, loadbasis, RadialGrid
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = eval_gto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([6, 100])

    Parameters
    ----------
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate
    to_transpose: book (default False)
        True for transposing the matrix.

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto

    """
    return evl("", wrapper, rgrid, to_transpose=to_transpose)


def eval_gradgto(wrapper: LibcintWrapper,
                 rgrid: torch.Tensor,
                 *,
                 to_transpose: bool = False) -> torch.Tensor:
    """Evaluates Grad GTO

    Examples
    --------
    >>> from deepchem.utils.dft_utils import eval_gradgto, AtomCGTOBasis, LibcintWrapper, loadbasis, RadialGrid
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = eval_gradgto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([3, 6, 100])

    Parameters
    ----------
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate
    to_transpose: book (default False)
        True for transposing the matrix.

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto
    """
    return evl("ip", wrapper, rgrid, to_transpose=to_transpose)


def eval_laplgto(wrapper: LibcintWrapper,
                 rgrid: torch.Tensor,
                 *,
                 to_transpose: bool = False) -> torch.Tensor:
    """Evaluates laplgto

    Examples
    --------
    >>> from deepchem.utils.dft_utils import eval_laplgto, AtomCGTOBasis, LibcintWrapper, loadbasis, RadialGrid
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = eval_laplgto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([6, 100])

    Parameters
    ----------
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate
    to_transpose: book (default False)
        True for transposing the matrix.

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto
    """
    return evl("lapl", wrapper, rgrid, to_transpose=to_transpose)


def pbc_eval_gto(wrapper: LibcintWrapper,
                 rgrid: torch.Tensor,
                 kpts: Optional[torch.Tensor] = None,
                 options: Optional[PBCIntOption] = None) -> torch.Tensor:
    """evaluate the basis in periodic boundary condition,
    i.e. evaluate sum_L exp(i*k*L) * phi(r - L)

    Examples
    --------
    >>> from deepchem.utils.dft_utils import pbc_eval_gto, AtomCGTOBasis, LibcintWrapper, loadbasis, Lattice, RadialGrid
    >>> dtype = torch.float64
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=dtype)
    >>> lattice = Lattice(a)
    >>> wrap = LibcintWrapper(atombases, True, lattice)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = pbc_eval_gto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([1, 6, 100])

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate. (ngrid, ndim)
    kpts: Optional[torch.Tensor] (default None)
        k-points in the Hamiltonian. (nkpts, ndim)
    options: Optional[PBCIntOption] (default None)

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto. (*ncomp, nkpts, nao, ngrid)

    """
    return pbc_evl("", wrapper, rgrid, kpts, options)


def pbc_eval_gradgto(wrapper: LibcintWrapper,
                     rgrid: torch.Tensor,
                     kpts: Optional[torch.Tensor] = None,
                     options: Optional[PBCIntOption] = None) -> torch.Tensor:
    """evaluate the basis in periodic boundary condition,
    i.e. evaluate sum_L exp(i*k*L) * phi(r - L)

    Examples
    --------
    >>> from deepchem.utils.dft_utils import pbc_eval_gradgto, AtomCGTOBasis, LibcintWrapper, loadbasis, Lattice, RadialGrid
    >>> dtype = torch.float64
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=dtype)
    >>> lattice = Lattice(a)
    >>> wrap = LibcintWrapper(atombases, True, lattice)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = pbc_eval_gradgto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([3, 1, 6, 100])

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate. (ngrid, ndim)
    kpts: Optional[torch.Tensor] (default None)
        k-points in the Hamiltonian. (nkpts, ndim)
    options: Optional[PBCIntOption] (default None)

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto. (*ncomp, nkpts, nao, ngrid)

    """
    return pbc_evl("ip", wrapper, rgrid, kpts, options)


def pbc_eval_laplgto(wrapper: LibcintWrapper,
                     rgrid: torch.Tensor,
                     kpts: Optional[torch.Tensor] = None,
                     options: Optional[PBCIntOption] = None) -> torch.Tensor:
    """evaluate the basis in periodic boundary condition,
    i.e. evaluate sum_L exp(i*k*L) * phi(r - L)

    Examples
    --------
    >>> from deepchem.utils.dft_utils import pbc_eval_laplgto, AtomCGTOBasis, LibcintWrapper, loadbasis, Lattice, RadialGrid
    >>> dtype = torch.float64
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]], dtype=dtype)
    >>> lattice = Lattice(a)
    >>> wrap = LibcintWrapper(atombases, True, lattice)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = pbc_eval_laplgto(wrap, grid.get_rgrid())
    >>> grad.shape
    torch.Size([1, 6, 100])

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    rgrid: torch.Tensor
        grid points position in the specified coordinate. (ngrid, ndim)
    kpts: Optional[torch.Tensor] (default None)
        k-points in the Hamiltonian. (nkpts, ndim)
    options: Optional[PBCIntOption] (default None)

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto. (*ncomp, nkpts, nao, ngrid)

    """
    return pbc_evl("lapl", wrapper, rgrid, kpts, options)


# pytorch function
class _EvalGTO(torch.autograd.Function):
    """wrapper class to provide the gradient for evaluating the contracted gto"""

    @staticmethod
    def forward(
            ctx,  # type: ignore
            coeffs: torch.Tensor,
            alphas: torch.Tensor,
            pos: torch.Tensor,
            rgrid: torch.Tensor,
            ao_to_atom: torch.Tensor,
            wrapper: LibcintWrapper,
            shortname: str,
            to_transpose: bool) -> torch.Tensor:
        """Forward function for EvalGTO.

        Parameters
        ----------
        coeffs: torch.Tensor
            The coefficients to get the orthogonal orbitals. (ngauss_tot)
        alphas: torch.Tensor
            gaussian exponents of the basis. (ngauss_tot)
        pos: torch.Tensor
            Position of the atom. (natom, ndim)
        rgrid: torch.Tensor
            Grid points positioned according to a center point. (ngrid, ndim)
        ao_to_atom: torch.Tensor
        (nao, ndim)
                wrapper: LibcintWrapper,
                shortname: str,
                to_transpose: bool) -> torch.Tensor:
        """

        res = gto_evaluator(wrapper, shortname, rgrid,
                            to_transpose)  # (*, nao, ngrid)
        ctx.save_for_backward(coeffs, alphas, pos, rgrid)
        ctx.other_info = (ao_to_atom, wrapper, shortname, to_transpose)
        return res

    @staticmethod
    def backward(  # type: ignore
            ctx, grad_res: torch.Tensor) -> Tuple[Optional[torch.Tensor], ...]:
        """Backward function for EvalGTO.

        Parameters
        ----------
        ctx : torch.autograd.Function
            Context object containing the saved tensors.
        grad_res : torch.Tensor
            Gradient of the result. (nao, ngrid)

        Returns
        -------
        Tuple[Optional[torch.Tensor], ...]
            Gradients for the coefficients, alphas, positions, rgrid

        """
        ao_to_atom, wrapper, shortname, to_transpose = ctx.other_info
        coeffs, alphas, pos, rgrid = ctx.saved_tensors

        if to_transpose:
            grad_res = grad_res.transpose(-2, -1)

        grad_alphas = None
        grad_coeffs = None
        if alphas.requires_grad or coeffs.requires_grad:
            u_wrapper, uao2ao = wrapper.get_uncontracted_wrapper()
            u_coeffs, u_alphas, u_pos = u_wrapper.params
            # (*, nu_ao, ngrid)
            u_grad_res = _gather_at_dims(grad_res, mapidxs=[uao2ao], dims=[-2])

            # get the scatter indices
            ao2shl = u_wrapper.ao_to_shell()

            # calculate the gradient w.r.t. coeffs
            if coeffs.requires_grad:
                grad_coeffs = torch.zeros_like(coeffs)  # (ngauss)

                # get the uncontracted version of the integral
                # (..., nu_ao, ngrid)
                dout_dcoeff = _EvalGTO.apply(*u_wrapper.params, rgrid,
                                             ao_to_atom, u_wrapper, shortname,
                                             False)

                # get the coefficients and spread it on the u_ao-length tensor
                coeffs_ao = torch.gather(coeffs, dim=-1,
                                         index=ao2shl)  # (nu_ao)
                dout_dcoeff = dout_dcoeff / coeffs_ao[:, None]
                grad_dcoeff = torch.einsum("...ur,...ur->u", u_grad_res,
                                           dout_dcoeff)  # (nu_ao)

                grad_coeffs.scatter_add_(dim=-1, index=ao2shl, src=grad_dcoeff)

            if alphas.requires_grad:
                grad_alphas = torch.zeros_like(alphas)

                new_sname = _get_evalgto_derivname(shortname, "a")
                # (..., nu_ao, ngrid)
                dout_dalpha = _EvalGTO.apply(*u_wrapper.params, rgrid,
                                             ao_to_atom, u_wrapper, new_sname,
                                             False)

                # _alphas_ao = torch.gather(alphas, dim=-1,
                #                           index=ao2shl)  # (nu_ao)
                grad_dalpha = -torch.einsum("...ur,...ur->u", u_grad_res,
                                            dout_dalpha)

                grad_alphas.scatter_add_(dim=-1, index=ao2shl, src=grad_dalpha)

        # calculate the gradient w.r.t. basis' pos and rgrid
        grad_pos = None
        grad_rgrid = None
        if rgrid.requires_grad or pos.requires_grad:
            opsname = _get_evalgto_derivname(shortname, "r")
            dresdr = _EvalGTO.apply(*ctx.saved_tensors, ao_to_atom, wrapper,
                                    opsname, False)  # (ndim, *, nao, ngrid)
            grad_r = dresdr * grad_res  # (ndim, *, nao, ngrid)

            if rgrid.requires_grad:
                grad_rgrid = grad_r.reshape(dresdr.shape[0], -1,
                                            dresdr.shape[-1])
                grad_rgrid = grad_rgrid.sum(dim=1).transpose(
                    -2, -1)  # (ngrid, ndim)

            if pos.requires_grad:
                grad_rao = torch.movedim(grad_r, -2, 0)  # (nao, ndim, *, ngrid)
                grad_rao = -grad_rao.reshape(*grad_rao.shape[:2], -1).sum(
                    dim=-1)  # (nao, ndim)
                grad_pos = torch.zeros_like(pos)  # (natom, ndim)
                grad_pos.scatter_add_(dim=0, index=ao_to_atom, src=grad_rao)

        return grad_coeffs, grad_alphas, grad_pos, grad_rgrid, \
            None, None, None, None, None, None


# evaluator (direct interfact to libcgto)
def gto_evaluator(wrapper: LibcintWrapper, shortname: str, rgrid: torch.Tensor,
                  to_transpose: bool):
    """Evaluate the contracted GTO

    Examples
    --------
    >>> from deepchem.utils.dft_utils import gto_evaluator, AtomCGTOBasis, LibcintWrapper, loadbasis, RadialGrid
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grad = gto_evaluator(wrap, "", grid.get_rgrid(), False)
    >>> grad.shape
    torch.Size([6, 100])

    Parameters
    ----------
    wrapper : LibcintWrapper
        Wrapper object containing the atomic information.
    shortname: str
        Short name of the integral.
    rgrid: torch.Tensor
        grid points position in the specified coordinate (ngrid, ndim)
    to_transpose: bool
        True for transposing the matrix.

    Returns
    -------
    torch.Tensor
        Gradient for evaluating the contracted gto.
        (*, nao, ngrid) if not to_transpose else (*, ngrid, nao)

    NOTE
    ----
    This function do not propagate gradient and should only be used
    in this file only

    """
    ngrid = rgrid.shape[0]
    nshells = len(wrapper)
    nao = wrapper.nao()
    opname = _get_evalgto_opname(shortname, wrapper.spherical)
    outshape = _get_evalgto_compshape(shortname) + (nao, ngrid)

    out = np.empty(outshape, dtype=np.float64)
    non0tab = np.ones(((ngrid + BLKSIZE - 1) // BLKSIZE, nshells),
                      dtype=np.int8)
    rgrid = rgrid.contiguous()
    coords = np.asarray(rgrid, dtype=np.float64, order='F')
    ao_loc = np.asarray(wrapper.full_shell_to_aoloc, dtype=np.int32)

    c_shls = (ctypes.c_int * 2)(*wrapper.shell_idxs)
    c_ngrid = ctypes.c_int(ngrid)

    # evaluate the orbital
    operator = getattr(CGTO(), opname)
    operator.restype = ctypes.c_double
    atm, bas, env = wrapper.atm_bas_env
    operator(c_ngrid, c_shls, np2ctypes(ao_loc), np2ctypes(out),
             np2ctypes(coords), np2ctypes(non0tab), np2ctypes(atm),
             int2ctypes(atm.shape[0]), np2ctypes(bas), int2ctypes(bas.shape[0]),
             np2ctypes(env))

    if to_transpose:
        out = np.ascontiguousarray(np.moveaxis(out, -1, -2))

    out_tensor = torch.as_tensor(out,
                                 dtype=wrapper.dtype,
                                 device=wrapper.device)
    return out_tensor


def _get_evalgto_opname(shortname: str, spherical: bool) -> str:
    """Returns the complete name of the evalgto operation

    Examples
    --------
    >>> from deepchem.utils.dft_utils.hamilton.intor.gtoeval import _get_evalgto_opname
    >>> _get_evalgto_opname("", False)
    'GTOval_cart'

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    spherical : bool
        True if the basis is spherical

    Returns
    -------
    str
        Name of the evalgto operation

    """
    sname = ("_" + shortname) if (shortname != "") else ""
    suffix = "_sph" if spherical else "_cart"
    return "GTOval%s%s" % (sname, suffix)


def _get_evalgto_compshape(shortname: str) -> Tuple[int, ...]:
    """returns the component shape of the evalgto function

    Examples
    --------
    >>> from deepchem.utils.dft_utils.hamilton.intor.gtoeval import _get_evalgto_compshape
    >>> _get_evalgto_compshape("ip")
    (3,)

    Parameters
    ----------
    shortname : str
        Short name of the integral.

    Returns
    -------
    Tuple[int, ...]
        Component shape of the evalgto function

    """

    # count "ip" only at the beginning
    n_ip = len(re.findall(r"^(?:ip)*(?:ip)?", shortname)[0]) // 2
    return (NDIM,) * n_ip


def _get_evalgto_derivname(shortname: str, derivmode: str):
    """Returns the derivative name of the evalgto operation

    Examples
    --------
    >>> from deepchem.utils.dft_utils.hamilton.intor.gtoeval import _get_evalgto_derivname
    >>> _get_evalgto_derivname("", "a")
    'rr'

    Parameters
    ----------
    shortname : str
        Short name of the integral.
    derivmode : str
        Derivative mode

    Returns
    -------
    str
        Derivative name of the evalgto operation

    """
    if derivmode == "r":
        return "ip%s" % shortname
    elif derivmode == "a":
        return "%srr" % shortname
    else:
        raise RuntimeError("Unknown derivmode: %s" % derivmode)



from transformers import BertForMaskedLM, BertForSequenceClassification, BertTokenizer, BertConfig, BertModel, PretrainedConfig
import torch.nn as nn
from deepchem.models.torch_models.hf_models import HuggingFaceModel
from typing import Union
import torch
from transformers.modeling_utils import PreTrainedModel


class ProtBERT(HuggingFaceModel):
    """
   ProtBERT model[1].


   ProtBERT model is based on BERT architecture and the current implementation
   supports only MLM pretraining and classification mode, as described by the
   authors in HuggingFace[2]. For classfication we currently only support
   Logistic regression and a simple Feed forward neural network.


   The model converts the input protein sequence into a vector through a trained BERT tokenizer, which is then
   processed by the corresponding model based on the task. `BertForMaskedLM` is used to facilitate the MLM
   pretraining task. For the sequence classification task, we follow `BertForSequenceClassification` but change
   the classifier to either a logistic regression (LogReg) or a feed-forward neural network (FFN), depending on
   the specified `cls_name`. The FFN is a simple 2-layer network with 512 as the hidden dimension.




   Examples
   --------
   >>> import os
   >>> import tempfile
   >>> tempdir = tempfile.mkdtemp()


   >>> # preparing dataset
   >>> import pandas as pd
   >>> import deepchem as dc
   >>> protein = ["MPCTTYLPLLLLLFLLPPPSVQSKV","SSGLFWMELLTQFVLTWPLVVIAFL"]
   >>> labels = [0,1]
   >>> df = pd.DataFrame(list(zip(protein, labels)), columns=["protein", "task1"])
   >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
   ...    df.to_csv(tmpfile.name)
   ...    loader = dc.data.CSVLoader(["task1"], feature_field="protein", featurizer=dc.feat.DummyFeaturizer())
   ...    dataset = loader.create_dataset(tmpfile.name)


   >>> # pretraining
   >>> from deepchem.models.torch_models.prot_bert import ProtBERT
   >>> pretrain_model_dir = os.path.join(tempdir, 'pretrain-model')
   >>> model_path = 'Rostlab/prot_bert'
   >>> pretrain_model = ProtBERT(task='mlm', HG_model_path=model_path, n_tasks=1, model_dir=pretrain_model_dir)  # mlm pretraining
   >>> pretraining_loss = pretrain_model.fit(dataset, nb_epoch=1)
   >>> del pretrain_model


   >>> finetune_model_dir = os.path.join(tempdir, 'finetune-model')
   >>> finetune_model = ProtBERT(task='classification', HG_model_path=model_path, n_tasks=1, model_dir=finetune_model_dir)
   >>> finetune_model.load_from_pretrained(pretrain_model_dir)
   >>> finetuning_loss = finetune_model.fit(dataset, nb_epoch=1)


   >>> # prediction and evaluation
   >>> result = finetune_model.predict(dataset)
   >>> eval_results = finetune_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.accuracy_score))


   >>> feat_extractor_model = ProtBERT(task='feature_extractor', HG_model_path=model_path, n_tasks=1, device = 'cpu')
   >>> protein = "M G L P V S W A P P A L W V L G C C A L L L S L W A"
   >>> tokenized_data = feat_extractor_model.tokenizer(protein,return_tensors='pt')
   >>> protbert_feats = feat_extractor_model.get_last_hidden_state(tokenized_data['input_ids'],tokenized_data['attention_mask'])


   References
   ----------
   .. [1] Elnaggar, Ahmed, et al. "Prottrans: Toward understanding the language of life through self-supervised learning." IEEE transactions on pattern analysis and machine intelligence 44.10 (2021): 7112-7127.
   .. [2] https://huggingface.co/Rostlab/prot_bert


   """

    def __init__(self,
                 task: str,
                 model_path: str = 'Rostlab/prot_bert',
                 n_tasks: int = 1,
                 cls_name: str = "LogReg",
                 classifier_net: Union[nn.Module, None] = None,
                 n_classes: int = 2,
                 **kwargs) -> None:
        """
       Parameters
       ----------
       task: str
           The task defines the type of learning task in the model. The supported tasks are
           - `mlm` - masked language modeling commonly used in pretraining
           - `classification` - use it for classification tasks
           - `feature_extractor` - use it along side the predict_embedding() method to extract features from a protein sequence
       model_path: str
           Path to the HuggingFace model
           - 'Rostlab/prot_bert' - Pretrained on Uniref100 dataset
           - `Rostlab/prot_bert_bfd` - Pretrained on BFD dataset
       n_tasks: int
           Number of prediction targets for a multitask learning model
       cls_name: str
           The classifier head to use for classification mode. Currently supports "FFN" and "LogReg" and custom classfier head.
       classifier_net: nn.Module, optional
           A custom classifier head to use for classification mode. The network must have input size of 1024.
       n_classes: int
           Number of classes for classification.
       """
        self.n_tasks: int = n_tasks
        tokenizer: BertTokenizer = BertTokenizer.from_pretrained(
            model_path, do_lower_case=False)
        protbert_config: PretrainedConfig = BertConfig.from_pretrained(
            pretrained_model_name_or_path=model_path,
            vocab_size=tokenizer.vocab_size)
        self.config = protbert_config
        model: PreTrainedModel
        if task == "mlm":
            model = BertForMaskedLM.from_pretrained(model_path)
        elif task == "classification":
            cls_head: Union[nn.Module, None]
            protbert_config.num_labels = n_classes
            if n_tasks == 1:
                protbert_config.problem_type = 'single_label_classification'
            else:
                protbert_config.problem_type = 'multi_label_classification'

            if (cls_name == "LogReg"):
                cls_head = nn.Linear(1024, n_classes)
            elif (cls_name == "FFN"):
                cls_head = nn.Sequential(nn.Linear(1024, 512), nn.ReLU(),
                                         nn.Linear(512, n_classes))
            elif (cls_name == "custom"):
                cls_head = classifier_net

            else:
                raise ValueError('Invalid classifier: {}.'.format(cls_name))

            model = BertForSequenceClassification.from_pretrained(
                model_path, config=protbert_config)
            assert cls_head is not None
            model.classifier = cls_head

        elif task == "feature_extractor":
            model = BertModel.from_pretrained(model_path,
                                              config=protbert_config)

        else:
            raise ValueError('Invalid task specification')
        super().__init__(model=model, task=task, tokenizer=tokenizer, **kwargs)

    def get_last_hidden_state(self, input_ids: torch.Tensor,
                              attention_mask: torch.Tensor) -> torch.Tensor:
        """
       Extracts the last hidden state from the model output.


       Parameters
       ----------
       input_ids : torch.Tensor
           Tensor containing tokenized input sequences.
       attention_mask : torch.Tensor
           Tensor indicating which positions should be attended to.


       Returns
       -------
       torch.Tensor
           The last hidden state of the model.
       """

        return self.model(input_ids, attention_mask).last_hidden_state[:, 0, :]



"""Part of this code adopted from https://github.com/diffqc/dqc"""
from typing import Optional, Dict, Any, Tuple, List, Union
import torch
from deepchem.utils.dft_utils import BaseSystem, SCF_QCCalc, BaseSCFEngine, SpinParam
from deepchem.utils.differentiation_utils import LinearOperator, lsymeig


class HF(SCF_QCCalc):
    """
    Performing Restricted or Unrestricted Kohn-Sham DFT calculation.

    Parameters
    ----------
    system: BaseSystem
        The system to be calculated.
    restricted: bool or None
        If True, performing restricted Kohn-Sham DFT. If False, it performs
        the unrestricted Kohn-Sham DFT.
        If None, it will choose True if the system is unpolarized and False if
        it is polarized
    variational: bool
        If True, then use optimization of the free orbital parameters to find
        the minimum energy.
        Otherwise, use self-consistent iterations.

    """

    def __init__(self,
                 system: BaseSystem,
                 restricted: Optional[bool] = None,
                 variational: bool = False):

        engine = HFEngine(system, restricted)
        super().__init__(engine, variational)


class HFEngine(BaseSCFEngine):
    """
    Engine to be used with Hartree Fock.
    This class provides the calculation of the self-consistency iteration step
    and the calculation of the post-calculation properties.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import HFEngine, BaseSystem, BaseHamilton, SpinParam, BaseGrid
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> from typing import List, Optional
    >>> class MyLinOp(LinearOperator):
    ...     def __init__(self, shape):
    ...         super(MyLinOp, self).__init__(shape)
    ...         self.param = torch.rand(shape)
    ...     def _getparamnames(self, prefix=""):
    ...         return [prefix + "param"]
    ...     def _mv(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmv(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _mm(self, x):
    ...         return torch.matmul(self.param, x)
    ...     def _rmm(self, x):
    ...         return torch.matmul(self.param.transpose(-2,-1).conj(), x)
    ...     def _fullmatrix(self):
    ...         return self.param
    >>> class MyHamilton(BaseHamilton):
    ...     def __init__(self):
    ...         self._nao = 2
    ...         self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
    ...         self._df = None
    ...     @property
    ...     def nao(self):
    ...         return self._nao
    ...     @property
    ...     def kpts(self):
    ...         return self._kpts
    ...     @property
    ...     def df(self):
    ...         return self._df
    ...     def build(self):
    ...         return self
    ...     def get_nuclattr(self):
    ...         return torch.ones((1, 1, self.nao, self.nao))
    ...     def get_e_elrep(self, dmtot):
    ...         return 2 * dmtot
    ...     def get_e_exchange(self, dm):
    ...         if isinstance(dm, SpinParam):
    ...             return SpinParam.sum(dm)
    ...         else:
    ...             return 2 * dm
    ...     def get_e_hcore(self, dm):
    ...         return 4 * dm
    ...     def get_elrep(self, dmtot):
    ...         return MyLinOp((self.nao + 1, self.nao + 1))
    ...     def get_exchange(self, dm):
    ...         return MyLinOp((self.nao + 1, self.nao + 1))
    ...     def get_kinnucl(self):
    ...         linop = MyLinOp((self.nao + 1, self.nao + 1))
    ...         return linop
    ...     def ao_orb2dm(self, orb: torch.Tensor,
    ...                   orb_weight: torch.Tensor) -> torch.Tensor:
    ...         return orb * orb_weight
    ...     def ao_orb_params2dm(
    ...         self,
    ...         ao_orb_params: torch.Tensor,
    ...         ao_orb_coeffs: torch.Tensor,
    ...         orb_weight: torch.Tensor,
    ...         with_penalty: Optional[float] = None) -> List[torch.Tensor]:
    ...         return [ao_orb_params * orb_weight, ao_orb_coeffs * orb_weight]
    >>> ham = MyHamilton()
    >>> class MySystem(BaseSystem):
    ...     def __init__(self):
    ...         self.hamiltonian = ham
    ...         self.grid = BaseGrid()
    ...     def get_hamiltonian(self):
    ...         return self.hamiltonian
    ...     def get_grid(self):
    ...         return self.grid
    ...     def requires_grid(self):
    ...         return True
    ...     def get_orbweight(self, polarized: bool = False) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
    ...         return SpinParam(torch.tensor([2.0]), torch.tensor([2.0]))
    ...     def get_nuclei_energy(self):
    ...         return torch.tensor(10.0)
    >>> system = MySystem()
    >>> engine = HFEngine(system, False)
    >>> engine.set_eigen_options(eigen_options={"method": "exacteig"})
    >>> engine.dm2energy(torch.tensor([2])).shape
    torch.Size([1])
    >>> engine.dm2scp(torch.tensor([2])).shape
    torch.Size([3, 3])
    >>> engine.scp2dm(torch.rand((2, 2, 2))).u.shape
    torch.Size([2, 1])

    """

    def __init__(self,
                 system: BaseSystem,
                 restricted: Optional[bool] = None,
                 build_grid_if_necessary: bool = False):

        # decide if this is restricted or not
        if restricted is None:
            self._polarized = bool(system.spin != 0)
        else:
            self._polarized = not restricted

        # construct the grid if the system requires it
        if build_grid_if_necessary and system.requires_grid():
            system.setup_grid()
            system.get_hamiltonian().setup_grid(system.get_grid())

        # build the basis
        self._hamilton = system.get_hamiltonian().build()
        self._system = system

        # get the orbital info
        self._orb_weight = system.get_orbweight(
            polarized=self._polarized)  # (norb,)
        self._norb = SpinParam.apply_fcn(
            lambda orb_weight: int(orb_weight.shape[-1]), self._orb_weight)

        # set up the 1-electron linear operator
        self._core1e_linop = self._hamilton.get_kinnucl()  # kinetic and nuclear

    def get_system(self) -> BaseSystem:
        """Return the system object.

        Returns
        -------
        BaseSystem
            The system object.

        """
        return self._system

    @property
    def shape(self):
        """Shape of the density matrix

        Returns
        -------
        Tuple[int, int]
            Shape of the density matrix.

        """
        return self._core1e_linop.shape

    @property
    def dtype(self):
        """Dtype of the density matrix

        Returns
        -------
        torch.dtype
            Dtype of the density matrix.

        """
        return self._core1e_linop.dtype

    @property
    def device(self):
        """Device of the density matrix

        Returns
        -------
        torch.device
            Device of the density matrix.

        """
        return self._core1e_linop.device

    @property
    def polarized(self):
        """Returns if the calculation is polarized

        Returns
        -------
        bool
            If the calculation is polarized.

        """
        return self._polarized

    def dm2scp(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Convert from density matrix to a self-consistent parameter (scp)

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix to be converted.

        Returns
        -------
        torch.Tensor
            Self-consistent parameter.

        """
        if isinstance(dm, torch.Tensor):  # unpolarized
            # scp is the fock matrix
            fork = self.__dm2fock(dm)
            assert isinstance(fork, LinearOperator), type(fork)
            return fork.fullmatrix()
        else:  # polarized
            # scp is the concatenated fock matrix
            fock = self.__dm2fock(dm)
            assert isinstance(fock, SpinParam), type(fock)
            mat_u = fock.u.fullmatrix().unsqueeze(0)
            mat_d = fock.d.fullmatrix().unsqueeze(0)
            return torch.cat((mat_u, mat_d), dim=0)

    def scp2dm(
            self,
            scp: torch.Tensor) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Convert from self-consistent parameter (scp) to density matrix

        Parameters
        ----------
        scp: torch.Tensor
            Self-consistent parameter to be converted.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        """
        if not self._polarized:
            fock = LinearOperator.m(_symm(scp), is_hermitian=True)
            return self.__fock2dm(fock)
        else:
            fock_u = LinearOperator.m(_symm(scp[0]), is_hermitian=True)
            fock_d = LinearOperator.m(_symm(scp[1]), is_hermitian=True)
            return self.__fock2dm(SpinParam(u=fock_u, d=fock_d))

    def scp2scp(self, scp: torch.Tensor) -> torch.Tensor:
        """Self-consistent iteration step from a self-consistent parameter (scp)
        to an scp

        Parameters
        ----------
        scp: torch.Tensor
            Self-consistent parameter to be converted.

        Returns
        -------
        torch.Tensor
            New self-consistent parameter.

        """
        dm = self.scp2dm(scp)
        return self.dm2scp(dm)

    def aoparams2ene(self,
                     aoparams: torch.Tensor,
                     aocoeffs: torch.Tensor,
                     with_penalty: Optional[float] = None) -> torch.Tensor:
        """Calculate the energy from the atomic orbital params

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor to be added to the energy.

        Returns
        -------
        torch.Tensor
            Energy value.

        """
        dm, penalty = self.aoparams2dm(aoparams, aocoeffs, with_penalty)
        ene = self.dm2energy(dm)
        return (ene + penalty) if penalty is not None else ene

    def aoparams2dm(self, aoparams: torch.Tensor, aocoeffs: torch.Tensor,
                    with_penalty: Optional[float] = None) -> \
            Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]:
        """Convert the aoparams to density matrix and penalty factor

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor to be added to the energy.

        Returns
        -------
        Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]
            Density matrix and the penalty factor.

        """
        aop = self.unpack_aoparams(aoparams)  # tensor or SpinParam of tensor
        aoc = self.unpack_aoparams(aocoeffs)  # tensor or SpinParam of tensor
        dm_penalty = SpinParam.apply_fcn(
            lambda aop, aoc, orb_weight: self._hamilton.ao_orb_params2dm(
                aop, aoc, orb_weight, with_penalty=with_penalty), aop, aoc,
            self._orb_weight)
        if with_penalty is not None:
            dm = SpinParam.apply_fcn(lambda dm_penalty: dm_penalty[0],
                                     dm_penalty)
            penalty: Optional[torch.Tensor] = SpinParam.sum(
                SpinParam.apply_fcn(lambda dm_penalty: dm_penalty[1],
                                    dm_penalty))
        else:
            dm = dm_penalty
            penalty = None
        return dm, penalty

    def pack_aoparams(
            self, aoparams: Union[torch.Tensor,
                                  SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Check if polarized, then pack it by concatenating them in the last dimension

        Parameters
        ----------
        aoparams: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Atomic orbital parameters.

        Returns
        -------
        torch.Tensor
            Packed atomic orbital parameters.

        """
        if isinstance(aoparams, SpinParam):
            return torch.cat((aoparams.u, aoparams.d), dim=-1)
        else:
            return aoparams

    def unpack_aoparams(
            self, aoparams: torch.Tensor
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Check if polarized, then construct the SpinParam (reverting the pack_aoparams)

        Parameters
        ----------
        aoparams: torch.Tensor
            Packed atomic orbital parameters.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Atomic orbital parameters.

        """
        if isinstance(self._norb, SpinParam):
            return SpinParam(u=aoparams[..., :self._norb.u],
                             d=aoparams[..., self._norb.u:])
        else:
            return aoparams

    def set_eigen_options(self, eigen_options: Dict[str, Any]) -> None:
        """Set the eigendecomposition (diagonalization) option

        Parameters
        ----------
        eigen_options: Dict[str, Any]
            Options for the eigendecomposition.

        """
        self.eigen_options = eigen_options

    def dm2energy(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Calculate the energy given the density matrix

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        torch.Tensor
            Energy.

        """
        dmtot = SpinParam.sum(dm)
        e_core = self._hamilton.get_e_hcore(dmtot)
        e_elrep = self._hamilton.get_e_elrep(dmtot)
        e_exch = self._hamilton.get_e_exchange(dm)
        return e_core + e_elrep + e_exch + self._system.get_nuclei_energy()

    def __dm2fock(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """From density matrix, returns the fock matrix

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            Fock matrix.

        """
        vhf = self.__dm2vhf(dm)
        fock = SpinParam.apply_fcn(lambda vhf: self._core1e_linop + vhf, vhf)
        return fock

    def __dm2vhf(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """from density matrix, returns the linear operator on electron-electron
        coulomb and exchange

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            Linear operator on electron-electron coulomb and exchange

        """
        elrep = self._hamilton.get_elrep(SpinParam.sum(dm))
        exch = self._hamilton.get_exchange(dm)
        vhf = SpinParam.apply_fcn(lambda exch: elrep + exch, exch)
        return vhf

    def __fock2dm(
        self, fock: Union[LinearOperator, SpinParam[LinearOperator]]
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Diagonalize the fock matrix and obtain the density matrix

        Parameters
        ----------
        fock: Union[LinearOperator, SpinParam[LinearOperator]]
            Fock matrix.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        """
        eigvals, eigvecs = self.diagonalize(fock, self._norb)
        dm = SpinParam.apply_fcn(
            lambda eivecs, orb_weights: self._hamilton.ao_orb2dm(
                eivecs, orb_weights), eigvecs, self._orb_weight)
        return dm

    def diagonalize(self, fock: Union[LinearOperator, SpinParam[LinearOperator]], norb: Union[int, SpinParam[int]]) -> \
            Union[Tuple[torch.Tensor, torch.Tensor], Tuple[SpinParam[torch.Tensor], SpinParam[torch.Tensor]]]:
        """Diagonalize the fock matrix

        Parameters
        ----------
        fock: Union[LinearOperator, SpinParam[LinearOperator]]
            Fock matrix.
        norb: Union[int, SpinParam[int]]
            Number of orbitals.

        Returns
        -------
        Union[Tuple[torch.Tensor, torch.Tensor], Tuple[SpinParam[torch.Tensor], SpinParam[torch.Tensor]]]
            Eigenvalues and eigenvectors.

        """
        ovlp = self._hamilton.get_overlap()
        if isinstance(fock, SpinParam):
            assert isinstance(self._norb, SpinParam), type(self._norb)
            assert isinstance(norb, SpinParam), type(norb)
            eivals_u, eivecs_u = lsymeig(A=fock.u,
                                         neig=norb.u,
                                         M=ovlp,
                                         **self.eigen_options)
            eivals_d, eivecs_d = lsymeig(A=fock.d,
                                         neig=norb.d,
                                         M=ovlp,
                                         **self.eigen_options)
            return SpinParam(u=eivals_u, d=eivals_d), SpinParam(u=eivecs_u,
                                                                d=eivecs_d)
        else:
            assert isinstance(norb, int)
            return lsymeig(A=fock, neig=norb, M=ovlp, **self.eigen_options)

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Parameter names for the given method

        Parameters
        ----------
        methodname: str
            Method name.
        prefix: str
            Prefix to be added to the parameter names.

        Returns
        -------
        List[str]
            Parameter names.

        """
        if methodname == "scp2scp":
            return self.getparamnames("scp2dm", prefix=prefix) + \
                self.getparamnames("dm2scp", prefix=prefix)
        elif methodname == "scp2dm":
            return self.getparamnames("__fock2dm", prefix=prefix)
        elif methodname == "dm2scp":
            return self.getparamnames("__dm2fock", prefix=prefix)
        elif methodname == "aoparams2ene":
            return self.getparamnames("aoparams2dm", prefix=prefix) + \
                self.getparamnames("dm2energy", prefix=prefix)
        elif methodname == "aoparams2dm":
            if isinstance(self._orb_weight, SpinParam):
                params = [prefix + "_orb_weight.u", prefix + "_orb_weight.d"]
            else:
                params = [prefix + "_orb_weight"]
            return params + \
                self._hamilton.getparamnames("ao_orb_params2dm", prefix=prefix + "_hamilton.")
        elif methodname == "pack_aoparams":
            return []
        elif methodname == "unpack_aoparams":
            return []
        elif methodname == "dm2energy":
            hprefix = prefix + "_hamilton."
            sprefix = prefix + "_system."
            return self._hamilton.getparamnames("get_e_hcore", prefix=hprefix) + \
                self._hamilton.getparamnames("get_e_elrep", prefix=hprefix) + \
                self._hamilton.getparamnames("get_e_exchange", prefix=hprefix) + \
                self._system.getparamnames("get_nuclei_energy", prefix=sprefix)
        elif methodname == "__fock2dm":
            if isinstance(self._orb_weight, SpinParam):
                params = [prefix + "_orb_weight.u", prefix + "_orb_weight.d"]
            else:
                params = [prefix + "_orb_weight"]
            return self.getparamnames("diagonalize", prefix=prefix) + \
                self._hamilton.getparamnames("ao_orb2dm", prefix=prefix + "_hamilton.") + \
                params
        elif methodname == "__dm2fock":
            return self._core1e_linop._getparamnames(prefix=prefix + "_core1e_linop.") + \
                self.getparamnames("__dm2vhf", prefix=prefix)
        elif methodname == "__dm2vhf":
            hprefix = prefix + "_hamilton."
            return self._hamilton.getparamnames("get_elrep", prefix=hprefix) + \
                self._hamilton.getparamnames("get_exchange", prefix=hprefix)
        elif methodname == "diagonalize":
            return self._hamilton.getparamnames("get_overlap",
                                                prefix=prefix + "_hamilton.")
        else:
            raise KeyError("Method %s has no paramnames set" % methodname)


def _symm(scp: torch.Tensor):
    """Forcely symmetrize the tensor

    Parameters
    ----------
    scp: torch.Tensor
        Tensor to be symmetrized.

    Returns
    -------
    torch.Tensor
        Symmetrized tensor.

    """
    return (scp + scp.transpose(-2, -1)) * 0.5



from deepchem.feat.vocabulary_builders.vocabulary_builder import VocabularyBuilder
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    import tokenizers
    import transformers


class HuggingFaceVocabularyBuilder(VocabularyBuilder):
    """Wrapper for building vocabulary from HuggingFace algorithms

    The wrapper class can be used to build huggingface vocabulary building algorithms
    with DeepChem.

    Parameters
    ----------
    model: tokenizers.models.Model
        The core tokenization algorithm to learn the vocabulary.
    trainers: tokenizers.trainers.Trainer
        The trainer capable of training the model

    Example
    -------
    >>> from tokenizers import models, trainers
    >>> from tokenizers.pre_tokenizers import Whitespace
    >>> model = models.BPE(unk_token="[UNK]")
    >>> special_tokens = ["[UNK]"]
    >>> trainer = trainers.BpeTrainer(vocab_size=20, special_tokens=special_tokens)
    >>> vb = HuggingFaceVocabularyBuilder(model=model, trainer=trainer)
    """

    def __init__(self, model: 'tokenizers.models.Model',
                 trainer: 'tokenizers.trainers.Trainer'):
        from tokenizers import Tokenizer
        self.model = model
        self.trainer = trainer
        self.tokenizer = Tokenizer(model)

    # superclass accepts a DeepChem dataset while huggingface vocabulary builders
    # reads data from file
    def build(self, paths: List[str]):  # type: ignore
        """Trains the algorithm on the vocabulary

        Parameters
        ----------
        paths: List[str]
            A list of file paths for training
        """
        self.tokenizer.train(paths, self.trainer)

    @classmethod
    def load(
        cls, fname: str
    ) -> 'transformers.tokenization_utils_fast.PreTrainedTokenizerFast':
        """Loads vocabulary from a vocabulary file

        Parameters
        ----------
        fname: str
            Filename to load vocabulary from
        """
        from transformers import PreTrainedTokenizerFast
        tokenizer = PreTrainedTokenizerFast(tokenizer_file=fname)
        return tokenizer  # type: ignore

    def save(self, fname: str) -> None:
        """Saves vocabulary to a file

        Parameters
        ----------
        fname: str
            A json file path
        """
        self.tokenizer.save(fname)



from __future__ import annotations
from abc import abstractmethod
import torch
import numpy as np
from typing import Union, Tuple
from deepchem.utils.dft_utils import BaseGrid


class RadialGrid(BaseGrid):
    """
    Grid for radially symmetric system. This grid consists grid_integrator
    and grid_transform specifiers.

    grid_integrator is to specify how to perform an integration on a fixed
    interval from -1 to 1.

    grid_transform is to transform the integration from the coordinate of
    grid_integrator to the actual coordinate.

    Examples
    --------
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> grid.get_rgrid().shape
    torch.Size([100, 1])
    >>> grid.get_dvolume().shape
    torch.Size([100])

    """

    def __init__(self,
                 ngrid: int,
                 grid_integrator: str = "chebyshev",
                 grid_transform: Union[str, BaseGridTransform] = "logm3",
                 dtype: torch.dtype = torch.float64,
                 device: torch.device = torch.device('cpu')):
        """Initialize the RadialGrid.

        Parameters
        ----------
        ngrid: int
            Number of grid points.
        grid_integrator: str (default "chebyshev")
            The grid integrator to use. Available options are "chebyshev",
            "chebyshev2", and "uniform".
        grid_transform: Union[str, BaseGridTransform] (default "logm3")
            The grid transformation to use. Available options are "logm3",
            "de2", and "treutlerm4".
        dtype: torch.dtype, optional (default torch.float64)
            The data type to use for the grid.
        device: torch.device, optional (default torch.device('cpu'))
            The device to use for the grid.

        """
        self._dtype = dtype
        self._device = device
        grid_transform_obj = get_grid_transform(grid_transform)

        # get the location and weights of the integration in its original
        # coordinate
        _x, _w = get_xw_integration(ngrid, grid_integrator)
        x = torch.as_tensor(_x, dtype=dtype, device=device)
        w = torch.as_tensor(_w, dtype=dtype, device=device)
        r = grid_transform_obj.x2r(x)  # (ngrid,)

        # get the coordinate in Cartesian
        r1 = r.unsqueeze(-1)  # (ngrid, 1)
        self.rgrid = r1

        # integration element
        drdx = grid_transform_obj.get_drdx(x)
        vol_elmt = 4 * np.pi * r * r  # (ngrid,)
        dr = drdx * w
        self.dvolume = vol_elmt * dr  # (ngrid,)

    @property
    def coord_type(self):
        """Returns the coordinate type of the grid.

        Returns
        -------
        str
            The coordinate type of the grid. For RadialGrid, this is "radial".

        """
        return "radial"

    @property
    def dtype(self):
        """Returns the data type of the grid.

        Returns
        -------
        torch.dtype
            The data type of the grid.

        """
        return self._dtype

    @property
    def device(self):
        """Returns the device of the grid.

        Returns
        -------
        torch.device
            The device of the grid.

        """
        return self._device

    def get_dvolume(self) -> torch.Tensor:
        """Returns the integration element of the grid.

        Returns
        -------
        torch.Tensor
            The integration element of the grid.

        """
        return self.dvolume

    def get_rgrid(self) -> torch.Tensor:
        """Returns the grid points.

        Returns
        -------
        torch.Tensor
            The grid points.

        """
        return self.rgrid

    def __getitem__(self, key: Union[int, slice]) -> RadialGrid:
        """Returns a sliced RadialGrid.

        Parameters
        ----------
        key: Union[int, slice]
            The index or slice to use for slicing the grid.

        Returns
        -------
        RadialGrid
            The sliced RadialGrid.

        """
        if isinstance(key, slice):
            return SlicedRadialGrid(self, key)
        else:
            raise KeyError("Indexing for RadialGrid is not defined")

    def getparamnames(self, methodname: str, prefix: str = ""):
        """Returns the parameter names for the given method.

        Parameters
        ----------
        methodname: str
            The name of the method.
        prefix: str, optional (default "")
            The prefix to use for the parameter names.

        Returns
        -------
        List[str]
            The parameter names for the given method.

        """
        if methodname == "get_dvolume":
            return [prefix + "dvolume"]
        elif methodname == "get_rgrid":
            return [prefix + "rgrid"]
        else:
            raise KeyError("getparamnames for %s is not set" % methodname)


def get_xw_integration(n: int, s0: str) -> Tuple[np.ndarray, np.ndarray]:
    """returns ``n`` points of integration from -1 to 1 and its integration
    weights

    Examples
    --------
    >>> x, w = get_xw_integration(100, "chebyshev")
    >>> x.shape
    (100,)
    >>> w.shape
    (100,)

    Parameters
    ----------
    n: int
        Number of grid points.
    s0: str
        The grid integrator to use. Available options are `chebyshev`,
        `chebyshev2`, and `uniform`.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The integration points and weights.

    References
    ----------
    .. [1] chebyshev polynomial eq (9) & (10) https://doi.org/10.1063/1.475719
    .. [2] Handbook of Mathematical Functions (Abramowitz & Stegun) p. 889

    """

    s = s0.lower()
    if s == "chebyshev":
        np1 = n + 1.
        icount = np.arange(n, 0, -1)
        ipn1 = icount * np.pi / np1
        sin_ipn1 = np.sin(ipn1)
        sin_ipn1_2 = sin_ipn1 * sin_ipn1
        xcheb = (np1 - 2 * icount) / np1 + 2 / np.pi * \
                (1 + 2. / 3 * sin_ipn1 * sin_ipn1) * np.cos(ipn1) * sin_ipn1
        wcheb = 16. / (3 * np1) * sin_ipn1_2 * sin_ipn1_2
        return xcheb, wcheb

    elif s == "chebyshev2":
        np1 = n + 1.0
        icount = np.arange(n, 0, -1)
        ipn1 = icount * np.pi / np1
        sin_ipn1 = np.sin(ipn1)
        xcheb = np.cos(ipn1)
        wcheb = np.pi / np1 * sin_ipn1
        return xcheb, wcheb

    elif s == "uniform":
        x = np.linspace(-1, 1, n)
        w = np.ones(n) * (x[1] - x[0])
        w[0] *= 0.5
        w[-1] *= 0.5
        return x, w
    else:
        avail = ["chebyshev", "chebyshev2", "uniform"]
        raise RuntimeError("Unknown grid_integrator: %s. Available: %s" %
                           (s0, avail))


class SlicedRadialGrid(RadialGrid):
    """Internal class to represent the sliced radial grid"""

    def __init__(self, obj: RadialGrid, key: slice):
        """Initialize the SlicedRadialGrid.

        Parameters
        ----------
        obj: RadialGrid
            The original RadialGrid.
        key: slice
            The slice to use for slicing the grid.

        """
        self._dtype = obj.dtype
        self._device = obj.device
        self.dvolume = obj.dvolume[key]
        self.rgrid = obj.rgrid[key]


# Grid Transformations


class BaseGridTransform(object):
    """Base class for grid transformation
    Grid transformation is to transform the integration from the coordinate of
    grid_integrator to the actual coordinate.

    It is used as a base class for other grid transformations.
    x2r and get_drdx are abstract methods that need to be implemented.

    """

    @abstractmethod
    def x2r(self, x: torch.Tensor) -> torch.Tensor:
        """Transform from x to r coordinate

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        r: torch.Tensor
            The coordinate from 0 to inf.

        """
        pass

    @abstractmethod
    def get_drdx(self, x: torch.Tensor) -> torch.Tensor:
        """Returns the dr/dx

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        drdx: torch.Tensor
            The dr/dx.

        """
        pass


class DE2Transformation(BaseGridTransform):
    """Double exponential formula grid transformation

    Examples
    --------
    >>> x = torch.linspace(-1, 1, 100)
    >>> r = DE2Transformation().x2r(x)
    >>> r.shape
    torch.Size([100])
    >>> drdx = DE2Transformation().get_drdx(x)
    >>> drdx.shape
    torch.Size([100])

    References
    ----------
    .. [1] eq (31) in https://link.springer.com/article/10.1007/s00214-011-0985-x

    """

    def __init__(self,
                 alpha: float = 1.0,
                 rmin: float = 1e-7,
                 rmax: float = 20):
        assert rmin < 1.0
        self.alpha = alpha
        self.xmin = -np.log(-np.log(rmin))  # approximate for small r
        self.xmax = np.log(rmax) / alpha  # approximate for large r

    def x2r(self, x: torch.Tensor) -> torch.Tensor:
        """Transform from x to r coordinate

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        r: torch.Tensor
            The coordinate from 0 to inf.

        """
        # xnew is from [xmin, xmax]
        xnew = 0.5 * (x * (self.xmax - self.xmin) + (self.xmax + self.xmin))
        # r is approximately from [rmin, rmax]
        r = torch.exp(self.alpha * xnew - torch.exp(-xnew))
        return r

    def get_drdx(self, x: torch.Tensor) -> torch.Tensor:
        """Returns the dr/dx

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        drdx: torch.Tensor
            The dr/dx.

        """
        r = self.x2r(x)
        xnew = 0.5 * (x * (self.xmax - self.xmin) + (self.xmax + self.xmin))
        return r * (self.alpha + torch.exp(-xnew)) * (0.5 *
                                                      (self.xmax - self.xmin))


class LogM3Transformation(BaseGridTransform):
    """LogM3 grid transformation

    Examples
    --------
    >>> x = torch.linspace(-1, 1, 100)
    >>> r = LogM3Transformation().x2r(x)
    >>> r.shape
    torch.Size([100])
    >>> drdx = LogM3Transformation().get_drdx(x)
    >>> drdx.shape
    torch.Size([100])

    References
    ----------
    .. [1] eq (12) in https://aip.scitation.org/doi/pdf/10.1063/1.475719

    """

    def __init__(self, ra: float = 1.0, eps: float = 1e-15):
        """Initialize the LogM3Transformation.

        Parameters
        ----------
        ra: float (default 1.0)
            The parameter to control the range of the grid.
        eps: float (default 1e-15)
            The parameter to avoid numerical instability.

        """
        self.ra = ra
        self.eps = eps
        self.ln2 = np.log(2.0 + eps)

    def x2r(self, x: torch.Tensor) -> torch.Tensor:
        """Transform from x to r coordinate

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        torch.Tensor
            The coordinate from 0 to inf.

        """
        return self.ra * (1 - torch.log1p(-x + self.eps) / self.ln2)

    def get_drdx(self, x: torch.Tensor) -> torch.Tensor:
        """Returns the dr/dx

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        torch.Tensor
            The dr/dx.

        """
        return self.ra / self.ln2 / (1 - x + self.eps)


class TreutlerM4Transformation(BaseGridTransform):
    """Treutler M4 grid transformation

    Examples
    --------
    >>> x = torch.linspace(-1, 1, 100)
    >>> r = TreutlerM4Transformation().x2r(x)
    >>> r.shape
    torch.Size([100])
    >>> drdx = TreutlerM4Transformation().get_drdx(x)
    >>> drdx.shape
    torch.Size([100])

    References
    ----------
    .. [1] eq (19) in https://doi.org/10.1063/1.469408

    """

    def __init__(self, xi: float = 1.0, alpha: float = 0.6, eps: float = 1e-15):
        """Initialize the TreutlerM4Transformation.

        Parameters
        ----------
        xi: float (default 1.0)
            The parameter to control the range of the grid.
        alpha: float (default 0.6)
            The parameter to control the range of the grid.
        eps: float (default 1e-15)
            The parameter to avoid numerical instability.

        """
        self._xi = xi
        self._alpha = alpha
        self._ln2 = np.log(2.0 + eps)
        self._eps = eps

    def x2r(self, x: torch.Tensor) -> torch.Tensor:
        """Transform from x to r coordinate

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        torch.Tensor
            The coordinate from 0 to inf.

        """
        a = 1.0 + self._eps
        r = self._xi / self._ln2 * (a + x) ** self._alpha * \
            (self._ln2 - torch.log1p(-x + self._eps))
        return r

    def get_drdx(self, x: torch.Tensor) -> torch.Tensor:
        """Returns the dr/dx

        Parameters
        ----------
        x: torch.Tensor
            The coordinate from -1 to 1.

        Returns
        -------
        torch.Tensor
            The dr/dx.

        """
        a = 1.0 + self._eps
        fac = self._xi / self._ln2 * (a + x)**self._alpha
        r1 = fac / (1 - x + self._eps)
        r2 = fac * self._alpha / (a + x) * (self._ln2 -
                                            torch.log1p(-x + self._eps))
        return r2 + r1


def get_grid_transform(s0: Union[str, BaseGridTransform]) -> BaseGridTransform:
    """grid transformation object from the input

    Examples
    --------
    >>> transform = get_grid_transform("logm3")
    >>> transform.x2r(torch.tensor([0.5]))
    tensor([2.])

    Parameters
    ----------
    s0: Union[str, BaseGridTransform]
        The grid transformation to use. Available options are `logm3`,
        `de2`, and `treutlerm4`.

    Returns
    -------
    BaseGridTransform
        The grid transformation object.

    Raises
    ------
    RuntimeError
        If the input is not a valid grid transformation.

    """
    if isinstance(s0, BaseGridTransform):
        return s0
    else:
        s = s0.lower()
        if s == "logm3":
            return LogM3Transformation()
        elif s == "de2":
            return DE2Transformation()
        elif s == "treutlerm4":
            return TreutlerM4Transformation()
        else:
            raise RuntimeError("Unknown grid transformation: %s" % s0)



from typing import Union, List

import deepchem as dc


class DistributedTrainer():
    r"""Perform distributed training of DeepChem models

    DistributedTrainer provides an interface for scaling the training of DeepChem
    model to multiple GPUs and nodes. To achieve this, it uses
    `Lightning <https://lightning.ai/>`_ under the hood. A DeepChem model
    is converted to a `LightningModule <https://lightning.ai/docs/pytorch/stable/common/lightning_module.html>`_
    using :class:`~deepchem.models.lightning.DCLightningModule` and a DeepChem dataset
    is converted to a PyTorch iterable dataset using :class:`~deepchem.models.lightning.DCLightningDatasetModule`.
    The model and dataset are then trained using Lightning `Trainer <https://lightning.ai/docs/pytorch/stable/common/trainer.html>`_.

    Example
    -------
    .. code-block:: python

        import deepchem as dc
        from deepchem.models.trainer import DistributedTrainer

        dataset = dc.data.DiskDataset('zinc100k')

        atom_vocab = GroverAtomVocabularyBuilder.load('zinc100k_atom_vocab.json')
        bond_vocab = GroverBondVocabularyBuilder.load('zinc100k_bond_vocab.json')

        model = GroverModel(task='pretraining',
                            mode='pretraining',
                            node_fdim=151,
                            edge_fdim=165,
                            features_dim=2048,
                            functional_group_size=85,
                            hidden_size=128,
                            learning_rate=0.0001,
                            batch_size=1,
                            dropout=0.1,
                            ffn_num_layers=2,
                            ffn_hidden_size=5,
                            num_attn_heads=8,
                            attn_hidden_size=128,
                            atom_vocab=atom_vocab,
                            bond_vocab=bond_vocab)

        trainer = DistributedTrainer(max_epochs=1,
                                     batch_size=64,
                                     num_workers=0,
                                     accelerator='gpu',
                                     distributed_strategy='ddp')
        trainer.fit(model, dataset)

    """

    def __init__(self,
                 max_epochs: int,
                 batch_size: int,
                 devices: Union[str, int, List[int]] = 'auto',
                 accelerator: str = 'auto',
                 distributed_strategy: str = 'auto'):
        """Initialise DistributedTrainer

        Parameters
        ----------
        max_epochs: int
            Maximum number of training epochs
        batch_size: int
            Batch size
        devices: str, optional (default: 'auto')
            Number of devices to train on (int) or which devices to train on (list or str).
        accelerator: str, optional (default: 'auto')
            Specify the accelerator to train the models (cpu, gpu, mps)
        distributed_strategy: str, (default: 'auto')
            Specify training strategy (ddp - distributed data parallel, fsdp - fully sharded data parallel, etc).
        When strategy is auto, the Trainer chooses the default strategy for the hardware (ddp for torch). Also, refer to Lightning's docs on strategy `here <https://lightning.ai/docs/pytorch/stable/extensions/strategy.html>`_ .
        """
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.devices = devices
        self.accelerator = accelerator
        self.distributed_strategy = distributed_strategy

    def fit(self, model: 'dc.models.Model', dataset: 'dc.data.Dataset') -> None:
        """Train the model using DistributedTrainer

        Parameters
        ----------
        model: dc.models.Model
            A deepchem model to train
        dataset: dc.data.DiskDataset
            A deepchem Dataset for training
        """
        import lightning as L
        from deepchem.models.lightning import DCLightningModule, DCLightningDatasetModule
        lit_model = DCLightningModule(model)
        lit_dataset = DCLightningDatasetModule(dataset,
                                               batch_size=self.batch_size)
        trainer = L.Trainer(max_epochs=self.max_epochs,
                            devices=self.devices,
                            accelerator=self.accelerator)
        trainer.fit(lit_model, lit_dataset)
        return
